""" Integration tests spanning WecOptTool.
"""
import pytest
from pytest import approx
import wecopttool as wot
import capytaine as cpy
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import Bounds


kplim = -1e1


@pytest.fixture(scope="module")
def f1():
    """Fundamental frequency [Hz]"""
    return 0.05


@pytest.fixture(scope="module")
def nfreq():
    """Number of frequencies in the array"""
    return 50

@pytest.fixture(scope='module')
def pto():
    """Basic PTO: unstructured, 1 DOF, mechanical power."""
    ndof = 1
    kinematics = np.eye(ndof)
    pto = wot.pto.PTO(ndof, kinematics)
    return pto


@pytest.fixture(scope='module')
def p_controller_pto():
    """Basic PTO: proportional (P) controller, 1 DOF, mechanical power."""
    ndof = 1
    pto = wot.pto.PTO(ndof=ndof, kinematics=np.eye(ndof),
                      controller=wot.pto.controller_p,
                      names=["P controller PTO"])
    return pto


@pytest.fixture(scope='module')
def pi_controller_pto():
    """Basic PTO: proportional-integral (PI) controller, 1 DOF, mechanical 
    power."""
    ndof = 1
    pto = wot.pto.PTO(ndof=ndof, kinematics=np.eye(ndof),
                      controller=wot.pto.controller_pi,
                      names=["PI controller PTO"])
    return pto


@pytest.fixture(scope="module")
def fb():
    """Capytaine FloatingBody object"""
    try:
        import wecopttool.geom as geom
    except ImportError:
        pytest.skip(
            'Skipping integration tests due to missing optional geometry ' +
            'dependencies. Run `pip install wecopttool[geometry]` to run ' +
            'these tests.'
            )
    mesh_size_factor = 0.5
    wb = geom.WaveBot()
    mesh = wb.mesh(mesh_size_factor)
    fb = cpy.FloatingBody.from_meshio(mesh, name="WaveBot")
    fb.add_translation_dof(name="Heave")
    return fb


@pytest.fixture(scope="module")
def bem(f1, nfreq, fb):
    """Boundary elemement model (Capytaine) results"""
    freq = wot.frequency(f1, nfreq, False)
    return wot.run_bem(fb, freq)


@pytest.fixture(scope='module')
def regular_wave(f1, nfreq):
    """Single frequency wave"""
    wfreq = 0.3
    wamp = 0.0625
    wphase = 0
    wdir = 0
    waves = wot.waves.regular_wave(f1, nfreq, wfreq, wamp, wphase, wdir)
    return waves


@pytest.fixture(scope='module')
def irregular_wave(f1, nfreq):
    """Idealized (Pierson-Moskowitz) spectrum wave"""
    freq = wot.frequency(f1, nfreq, False)
    fp = 0.3
    hs = 0.0625*1.9
    spec = wot.waves.pierson_moskowitz_spectrum(freq, fp, hs)
    waves = wot.waves.long_crested_wave(spec)
    return waves


@pytest.fixture(scope='module')
def wec_from_bem(f1, nfreq, bem, fb, pto):
    """Simple WEC: 1 DOF, no constraints."""
    mass = wot.hydrostatics.inertia_matrix(fb).values
    hstiff = wot.hydrostatics.stiffness_matrix(fb).values
    f_add = {"PTO": pto.force_on_wec}
    wec = wot.WEC.from_bem(bem, mass, hstiff, f_add=f_add)
    return wec


@pytest.fixture(scope='module')
def wec_from_floatingbody(f1, nfreq, fb, pto):
    """Simple WEC: 1 DOF, no constraints."""
    mass = wot.hydrostatics.inertia_matrix(fb).values
    hstiff = wot.hydrostatics.stiffness_matrix(fb).values
    f_add = {"PTO": pto.force_on_wec}
    wec = wot.WEC.from_floating_body(fb, f1, nfreq, mass, hstiff, f_add=f_add)
    return wec


@pytest.fixture(scope='module')
def wec_from_impedance(bem, pto, fb):
    """Simple WEC: 1 DOF, no constraints."""
    bemc = bem.copy().transpose(
        "radiating_dof", "influenced_dof", "omega", "wave_direction")
    omega = bemc['omega'].values
    w = np.expand_dims(omega, [0, 1])
    A = bemc['added_mass'].values
    B = bemc['radiation_damping'].values
    mass = wot.hydrostatics.inertia_matrix(fb).values
    hstiff = wot.hydrostatics.stiffness_matrix(fb).values
    K = np.expand_dims(hstiff, 2)

    freqs = omega / (2 * np.pi)
    impedance = (A + mass)*(1j*w) + B + K/(1j*w)
    exc_coeff = bem['Froude_Krylov_force'] + bem['diffraction_force']
    f_add = {"PTO": pto.force_on_wec}

    wec = wot.WEC.from_impedance(freqs, impedance, exc_coeff, hstiff, f_add)
    return wec


@pytest.fixture(scope='module')
def resonant_wave(f1, nfreq, fb, bem):
    """Regular wave at natural frequency of the WEC"""
    mass = wot.hydrostatics.inertia_matrix(fb).values
    hstiff = wot.hydrostatics.stiffness_matrix(fb).values
    hd = wot.linear_hydrodynamics(bem, mass, hstiff)
    Zi = wot.hydrodynamic_impedance(hd)
    wn = Zi['omega'][np.abs(Zi).argmin(dim='omega')].item()
    waves = wot.waves.regular_wave(f1, nfreq, freq=wn/2/np.pi, amplitude=0.1)
    return waves


def test_solve_callback(wec_from_bem, regular_wave, pto, nfreq, capfd):
    """Check that user can set a custom callback"""

    cbstring = 'hello world!'

    def my_callback(my_wec, x_wec, x_opt, wave):
        print(cbstring)

    _ = wec_from_bem.solve(regular_wave,
                           obj_fun=pto.average_power,
                           nstate_opt=2*nfreq,
                           scale_x_wec=1.0,
                           scale_x_opt=0.01,
                           scale_obj=1e-1,
                           callback=my_callback,
                           optim_options={'maxiter': 1})

    out, err = capfd.readouterr()

    assert out.split('\n')[0] == cbstring


@pytest.mark.parametrize("bounds_opt",
                         [Bounds(lb=kplim, ub=0), ((kplim, 0),)])
def test_solve_bounds(bounds_opt, wec_from_bem, regular_wave,
                      p_controller_pto):
    """Confirm that bounds are not violated and scale correctly when 
    passing bounds argument as both as Bounds object and a tuple"""

    # replace unstructured controller with propotional controller
    wec_from_bem.forces['PTO'] = p_controller_pto.force_on_wec

    res = wec_from_bem.solve(waves=regular_wave,
                             obj_fun=p_controller_pto.average_power,
                             nstate_opt=1,
                             x_opt_0=[kplim*0.1],
                             optim_options={'maxiter': 2e1,
                                            'ftol': 1e-8},
                             bounds_opt=bounds_opt,
                             )

    assert pytest.approx(kplim, 1e-5) == res['x'][-1]


def test_same_wec_init(wec_from_bem,
                       wec_from_floatingbody,
                       wec_from_impedance,
                       f1,
                       nfreq):
    """Test that different init methods for WEC class produce the same object
    """

    waves = wot.waves.regular_wave(f1, nfreq, 0.3, 0.0625)
    np.random.seed(0)
    x_wec_0 = np.random.randn(wec_from_bem.nstate_wec)
    np.random.seed(1)
    x_opt_0 = np.random.randn(wec_from_bem.nstate_wec)
    bem_res = wec_from_bem._resid_fun(x_wec_0, x_opt_0, waves)
    fb_res = wec_from_floatingbody._resid_fun(x_wec_0, x_opt_0, waves)
    imp_res = wec_from_impedance._resid_fun(x_wec_0, x_opt_0, waves)

    assert fb_res == approx(bem_res, rel=0.01)
    assert imp_res == approx(bem_res, rel=0.01)


class TestTheoreticalPowerLimits:
    """Compare power from numerical solutions against known theoretical limits
    """

    @pytest.fixture(scope='class')
    def mass(self, fb):
        """Rigid-body mass"""
        return wot.hydrostatics.inertia_matrix(fb).values

    @pytest.fixture(scope='class')
    def hstiff(self, fb):
        """Hydrostatic stiffness"""
        return wot.hydrostatics.stiffness_matrix(fb).values

    @pytest.fixture(scope='class')
    def hydro_impedance(self, bem, mass, hstiff):
        """Intrinsic hydrodynamic impedance"""
        hd = wot.linear_hydrodynamics(bem, mass, hstiff)
        hd = wot.check_linear_damping(hd)
        Zi = wot.hydrodynamic_impedance(hd)
        return Zi

    def test_p_controller_resonant_wave(self,
                                        bem,
                                        resonant_wave,
                                        p_controller_pto,
                                        mass,
                                        hstiff,
                                        hydro_impedance):
        """Proportional controller should match optimum for natural resonant 
        wave"""

        f_add = {"PTO": p_controller_pto.force_on_wec}
        wec = wot.WEC.from_bem(bem, mass, hstiff, f_add=f_add)

        res = wec.solve(waves=resonant_wave,
                        obj_fun=p_controller_pto.average_power,
                        nstate_opt=1,
                        x_wec_0=1e-1*np.ones(wec.nstate_wec),
                        x_opt_0=[-1470],
                        scale_x_wec=1e2,
                        scale_x_opt=1e-3,
                        scale_obj=1e-1,
                        optim_options={'ftol': 1e-10},
                        bounds_opt=((-1*np.infty, 0),),
                        )

        power_sol = -1*res['fun']

        res_fd, _ = wec.post_process(res, resonant_wave,nsubsteps=1)
        Fex = res_fd.force.sel(
            type=['Froude_Krylov', 'diffraction']).sum('type')
        power_optimal = (np.abs(Fex)**2/8 / np.real(hydro_impedance.squeeze())
                         ).squeeze().sum('omega').item()

        assert power_sol == approx(power_optimal, rel=0.02)

    def test_pi_controller_regular_wave(self,
                                        bem,
                                        regular_wave,
                                        pi_controller_pto,
                                        mass,
                                        hstiff,
                                        hydro_impedance):
        """PI controller matches optimal for any regular wave"""

        f_add = {"PTO": pi_controller_pto.force_on_wec}
        wec = wot.WEC.from_bem(bem, mass, hstiff, f_add=f_add)

        res = wec.solve(waves=regular_wave,
                        obj_fun=pi_controller_pto.average_power,
                        nstate_opt=2,
                        x_wec_0=1e-1*np.ones(wec.nstate_wec),
                        x_opt_0=[-1e3, 1e4],
                        scale_x_wec=1e2,
                        scale_x_opt=1e-3,
                        scale_obj=1e-2,
                        optim_options={'maxiter': 50},
                        bounds_opt=((-1e4, 0), (0, 2e4),)
                        )

        power_sol = -1*res['fun']

        res_fd, _ = wec.post_process(res, regular_wave, nsubsteps=1)
        Fex = res_fd.force.sel(
            type=['Froude_Krylov', 'diffraction']).sum('type')
        power_optimal = (np.abs(Fex)**2/8 / np.real(hydro_impedance.squeeze())
                         ).squeeze().sum('omega').item()

        assert power_sol == approx(power_optimal, rel=1e-4)

    def test_unstructured_controller_irregular_wave(self,
                                                    fb,
                                                    bem,
                                                    regular_wave,
                                                    pto,
                                                    nfreq,
                                                    mass,
                                                    hstiff,
                                                    hydro_impedance):
        """Unstructured (numerical optimal) controller matches optimal for any 
        irregular wave when unconstrained"""

        f_add = {"PTO": pto.force_on_wec}
        wec = wot.WEC.from_bem(bem, mass, hstiff, f_add=f_add)

        res = wec.solve(waves=regular_wave,
                        obj_fun=pto.average_power,
                        nstate_opt=2*nfreq,
                        x_wec_0=1e-1*np.ones(wec.nstate_wec),
                        scale_x_wec=1e2,
                        scale_x_opt=1e-2,
                        scale_obj=1e-2,
                        )

        power_sol = -1*res['fun']

        res_fd, _ = wec.post_process(res, regular_wave, nsubsteps=1)
        Fex = res_fd.force.sel(
            type=['Froude_Krylov', 'diffraction']).sum('type')
        power_optimal = (np.abs(Fex)**2/8 / np.real(hydro_impedance.squeeze())
                         ).squeeze().sum('omega').item()

        assert power_sol == approx(power_optimal, rel=1e-3)



""" Unit tests for functions in the :python:`core.py` module.
Does not include the :python:`WEC` class.
"""

import os
import random

import pytest
from pytest import approx
import numpy as np
import xarray as xr
import capytaine as cpy

import wecopttool as wot


# small problem setup with synthetic BEM data
@pytest.fixture(scope='module')
def f1():
    """Fundamental frequency [Hz]."""
    return 0.12


@pytest.fixture(scope='module')
def nfreq():
    """Number of frequencies in frequency vector."""
    return 5


@pytest.fixture(scope='module')
def nsubsteps():
    """Number of sub-steps between default step size for the time
    vector.
    """
    return random.randint(2, 10)


@pytest.fixture(scope="module")
def ncomponents(nfreq):
    """Number of components in the WEC state."""
    return wot.ncomponents(nfreq)


@pytest.fixture(scope='module')
def bem_data(f1, nfreq):
    """Synthetic BEM data."""
    # TODO - start using single BEM solution across entire test suite
    coords = {
        'omega': [2*np.pi*(ifreq+1)*f1 for ifreq in range(nfreq)],
        'influenced_dof': ['DOF_1', 'DOF_2'],
        'radiating_dof': ['DOF_1', 'DOF_2'],
        'wave_direction': [0.0, 1.5, 2.1],
    }

    ndof = 2; ndir = 3;
    radiation_dims = ['radiating_dof', 'influenced_dof', 'omega']
    excitation_dims = ['influenced_dof', 'wave_direction', 'omega']

    added_mass = np.ones([ndof, ndof, nfreq])
    radiation_damping = np.ones([ndof, ndof, nfreq])
    diffraction_force = np.ones([ndof, ndir, nfreq], dtype=complex) + 1j
    Froude_Krylov_force = np.ones([ndof, ndir, nfreq], dtype=complex) + 1j

    data_vars = {
        'added_mass': (radiation_dims, added_mass),
        'radiation_damping': (radiation_dims, radiation_damping),
        'diffraction_force': (excitation_dims, diffraction_force),
        'Froude_Krylov_force': (excitation_dims, Froude_Krylov_force)
    }
    return xr.Dataset(data_vars=data_vars, coords=coords)


@pytest.fixture(scope='module')
def hydro_data(bem_data):
    """Synthetic hydro-data containing inertia, stiffness, and friction
    in addition to the coefficients in `bem_data`."""
    ndof = len(bem_data.influenced_dof)
    inertia_matrix = np.ones([ndof, ndof])
    stiffness = np.ones([ndof, ndof])
    friction = np.ones([ndof, ndof])
    data = wot.linear_hydrodynamics(
        bem_data, inertia_matrix, stiffness, friction
    )
    return data


# fixture: regular wave
@pytest.fixture(scope="module")
def wave_regular(f1, nfreq):
    """Wave structure consisting of a single regular wave."""
    n = np.random.randint(1, nfreq)
    freq = n*f1  # Hz
    amp = 1.1  # m
    phase = 24.5  # degrees
    wave = wot.waves.regular_wave(f1, nfreq, freq, amp, phase)
    params = {'n': n, 'amp': amp, 'phase': phase}
    return wave, params


@pytest.fixture(scope="module")
def ndof_waves():
    """Number of WEC degrees of freedom to consider for the wave tests.
    """
    return 3


@pytest.fixture(scope="module")
def exc_coeff(wave_regular, waves_multi, ndof_waves, nfreq, f1):
    """Excitation coefficients for the WEC."""
    # regular wave parameter
    wave_regular, params_reg = wave_regular
    n_r = params_reg['n']
    # multi-directional wave parameters
    waves_multi, params_multi = waves_multi
    directions = np.deg2rad(params_multi['directions'])
    ndir = len(directions)
    n_m = params_multi['n']
    # excitation force coefficients
    exc_coeff = np.zeros([nfreq, ndir, ndof_waves], dtype=complex)
    exc_coeff[n_m-1, 0, :] = [1+11j, 2+22j, 3+33j]
    exc_coeff[n_m-1, 1, :] = [4+44j, 5+55j, 6+66j]
    exc_coeff[n_r-1, 0, :] = [1+11j, 2+22j, 3+33j]
    exc_coeff[n_r-1, 1, :] = [4+44j, 5+55j, 6+66j]
    omega = 2*np.pi*wot.frequency(f1, nfreq, False)
    coords = {
        'omega': (['omega'], omega, {'units': 'rad/s'}),
        'wave_direction': (['wave_direction'], directions, {'units': 'rad'}),
        'influenced_dof': (['influenced_dof'], ['DOF_1', 'DOF_2', 'DOF_3'], {})
        }
    return xr.DataArray(exc_coeff, coords=coords, attrs={})


@pytest.fixture(scope="module")
def fexc_regular(nfreq, wave_regular, exc_coeff, ndof_waves):
    """Excitation force for the regular wave."""
    # wave elevation
    wave_regular, params_reg = wave_regular
    n_r = params_reg['n']
    wave_elev_regular = (
        params_reg['amp'] *
        np.exp(1j*wot.degrees_to_radians(params_reg['phase'])))
    # excitation force
    fexc = np.zeros([nfreq, ndof_waves], dtype=complex)
    for idof in range(ndof_waves):
        idir = 0
        fexc[n_r-1, idof] = (
            exc_coeff[n_r-1, idir, idof] * wave_elev_regular)
    return fexc


# fixture: sea-state composed of two wave components with different
# frequencies and directions.
@pytest.fixture(scope="module")
def waves_multi(f1, nfreq):
    """Waves structure composed of two sinusoidal waves in different
    directions.
    """
    n = np.random.randint(1, nfreq)
    directions = [0.0, 30.0]
    ndir = len(directions)

    amplitudes = np.zeros([nfreq, ndir])
    phases = np.zeros([nfreq, ndir])
    amp0, amp1 = 1.2, 2.1
    phase0, phase1 = 26, -13
    amplitudes[n-1, :] = [amp0, amp1]
    phases[n-1, :] = [phase0, phase1]

    waves = wot.waves.elevation_fd(f1, nfreq, directions, amplitudes, phases)

    params = {'n': n, 'directions': directions, 'amp0': amp0, 'amp1': amp1,
              'phase0': phase0, 'phase1': phase1}

    return waves, params


@pytest.fixture(scope="module")
def fexc_multi(nfreq, waves_multi, exc_coeff, ndof_waves):
    """Excitation force for the multi-directional wave."""
    # wave elevation
    waves_multi, params_multi = waves_multi
    n_m = params_multi['n']
    directions = np.deg2rad(params_multi['directions'])
    ndir = len(directions)
    waves_elev_multi = [
        params_multi['amp0'] * np.exp(1j*np.deg2rad(params_multi['phase0'])),
        params_multi['amp1'] * np.exp(1j*np.deg2rad(params_multi['phase1']))]
    # excitation force
    fexc = np.zeros([nfreq, ndof_waves], dtype=complex)
    for idof in range(ndof_waves):
        for idir in range(ndir):
            fexc[n_m-1, idof] += \
                exc_coeff[n_m-1, idir, idof] * waves_elev_multi[idir]
    return fexc


# fixture: synthetic impedance
@pytest.fixture(scope="module")
def ndof_imp():
    """Number of degrees of freedom in synthetic impedance."""
    return 2


@pytest.fixture(scope="module")
def nfreq_imp():
    """Number of frequencies in synthetic impedance."""
    return 2


@pytest.fixture(scope="module")
def rao(ndof_imp, nfreq_imp):
    """Synthetic RAO transfer matrix."""
    rao = np.empty([ndof_imp, ndof_imp, nfreq_imp], dtype=complex)
    rao[0, 0, :] = [0+1j, 0+2j]
    rao[1, 0, :] = [1+1j, 11+2j]
    rao[0, 1, :] = [2+1j, 22+2j]
    rao[1, 1, :] = [3+1j, 33+2j]
    return rao


@pytest.fixture(scope="module")
def mimo(nfreq_imp):
    """Correct MIMO matrix corresponding to the synthetic RAO
    transfer matrix.
    """
    ncomponents = wot.ncomponents(nfreq_imp)
    mimo = np.zeros([ncomponents*2, ncomponents*2])
    mimo[:ncomponents, :ncomponents] = np.array([
        [0, 0,  0, 0], #  0
        [0, 0, -1, 0], #  0
        [0, 1,  0, 0], #  0
        [0, 0,  0, 0], # -2
    #   [0, 0,  0, 2], #  0
    ])
    mimo[ncomponents:, :ncomponents] = np.array([
        [0, 0,  0,  0], #  0
        [0, 1, -1,  0], #  0
        [0, 1,  1,  0], #  0
        [0, 0,  0, 11], # -2
    #   [0, 0,  0,  2], # 11
    ])
    mimo[:ncomponents, ncomponents:] = np.array([
        [0, 0,  0,  0], #  0
        [0, 2, -1,  0], #  0
        [0, 1,  2,  0], #  0
        [0, 0,  0, 22], # -2
    #   [0, 0,  0,  2], # 22
    ])
    mimo[ncomponents:, ncomponents:] = np.array([
        [0, 0,  0,  0], #  0
        [0, 3, -1,  0], #  0
        [0, 1,  3,  0], #  0
        [0, 0,  0, 33], # -2
    #   [0, 0,  0,  2], # 33
    ])
    return mimo


class TestNComponents:
    """Test function :python:`ncomponents`."""

    def test_default(self, ncomponents, nfreq):
        """Test default, which will include a zero-frequency component.
        """
        assert ncomponents == 2*nfreq

    def test_nozero(self, nfreq):
        """Test without a zero-frequency component."""
        assert wot.ncomponents(nfreq, False) == 2*nfreq - 1


class TestFrequency:
    """Test function :python:`frequency`."""

    @pytest.fixture(scope="class")
    def freqs(self, f1, nfreq):
        """Frequency vector including the zero-frequency (default)."""
        return wot.frequency(f1, nfreq)

    @pytest.fixture(scope="class")
    def freqs_nm(self, f1, nfreq):
        """Frequency vector without the zero-frequency (mean)."""
        return wot.frequency(f1, nfreq, False)

    def test_shape(self, freqs, nfreq):
        """Test that the default frequency vector has the correct
        shape.
        """
        assert ((freqs.ndim==1) and (len(freqs)==nfreq+1))

    def test_first_last(self, freqs, nfreq, f1):
        """Test the first and last frequencies in the default vector."""
        assert ((freqs[0]==0.0) and (freqs[-1]==approx(f1*nfreq)))

    def test_evenly_spaced(self, freqs):
        """Test that the default frequency vector is evenly-spaced."""
        assert np.diff(freqs)==approx(np.diff(freqs)[0])

    def test_shape_nm(self, freqs_nm, nfreq):
        """Test that the "no-mean" frequency vector has the correct
        shape.
        """
        assert ((freqs_nm.ndim==1) and (len(freqs_nm)==nfreq))

    def test_first_last_nm(self, freqs_nm, nfreq, f1):
        """Test the first and last frequencies in the "no-mean" vector.
        """
        assert ((freqs_nm[0]==approx(f1)) and (freqs_nm[-1]==approx(f1*nfreq)))

    def test_evenly_spaced_nm(self, freqs_nm):
        """Test that the "no-mean" frequency vector is evenly-spaced."""
        assert np.diff(freqs_nm)==approx(np.diff(freqs_nm)[0])


class TestTime:
    """Test function :python:`time`."""

    @pytest.fixture(scope="class")
    def time(self, f1, nfreq):
        """Time vector with default :python:`nsubstep=1`."""
        return wot.time(f1, nfreq)

    @pytest.fixture(scope="class")
    def time_sub(self, f1, nfreq, nsubsteps):
        """Time vector with sub-steps."""
        return wot.time(f1, nfreq, nsubsteps)

    def test_shape(self, time, ncomponents):
        """Test that the default time vector has the correct shape."""
        t = time
        n = 1
        assert ((t.ndim==1) and (len(t)==ncomponents*n))

    def test_first_last(self, time, f1):
        """Test the first and last time steps in the default vector."""
        t = time
        t_last = 1.0/f1-np.diff(t)[0]
        assert ((t[0]==0.0) and (t[-1]==approx(t_last)))

    def test_evenly_spaced(self, time):
        """Test that the default time vector is evenly-spaced."""
        t = time
        assert np.diff(t)==approx(np.diff(t)[0])

    def test_shape_sub(self, time_sub, ncomponents, nsubsteps):
        """Test that the time vector with sub-steps has the correct
        shape.
        """
        t = time_sub
        n = nsubsteps
        assert ((t.ndim==1) and (len(t)==ncomponents*n))

    def test_first_last_sub(self, time_sub, f1):
        """Test the first and last time steps in the vector with
        sub-stpes.
        """
        t = time_sub
        t_last = 1.0/f1-np.diff(t)[0]
        assert ((t[0]==0.0) and (t[-1]==approx(t_last)))

    def test_evenly_spaced_sub(self, time_sub):
        """Test that the time vector with sub-steps is evenly-spaced."""
        t = time_sub
        assert np.diff(t)==approx(np.diff(t)[0])


class TestTimeMat:
    """Test function :python:`time_mat`."""

    @pytest.fixture(scope="class")
    def f1_tm(self,):
        """Fundamental frequency [Hz] for the synthetic time matrix."""
        return 0.5

    @pytest.fixture(scope="class")
    def nfreq_tm(self,):
        """Number of frequencies (harmonics) for the synthetic time
        matrix.
        """
        return 2

    @pytest.fixture(scope="class")
    def time_mat(self, f1_tm, nfreq_tm):
        """Correct/expected time matrix."""
        f = np.array([0, 1, 2])*f1_tm
        w = 2*np.pi * f
        t = 1/(2*nfreq_tm) * 1/f1_tm * np.arange(0, 2*nfreq_tm)
        c, s = np.cos, np.sin
        mat = np.array([
            [1,            1,             0,            1],
            [1, c(w[1]*t[1]), -s(w[1]*t[1]), c(w[2]*t[1])],
            [1, c(w[1]*t[2]), -s(w[1]*t[2]), c(w[2]*t[2])],
            [1, c(w[1]*t[3]), -s(w[1]*t[3]), c(w[2]*t[3])],
        ])
        return mat

    @pytest.fixture(scope="class")
    def time_mat_sub(self, f1, nfreq, nsubsteps):
        """Time matrix with sub-steps."""
        return wot.time_mat(f1, nfreq, nsubsteps)

    def test_time_mat(self, time_mat, f1_tm, nfreq_tm):
        """Test the default created time matrix."""
        calculated = wot.time_mat(f1_tm, nfreq_tm)
        assert calculated==approx(time_mat)

    def test_shape(self, time_mat_sub, ncomponents, nsubsteps):
        """Test the shape of the time matrix with sub-steps."""
        assert time_mat_sub.shape==(nsubsteps*ncomponents, ncomponents)

    def test_zero_freq(self, time_mat_sub):
        """Test the zero-frequency components of the time matrix with
        sub-steps.
        """
        assert all(time_mat_sub[:, 0]==1.0)

    def test_time_zero(self, time_mat_sub, nfreq):
        """Test the components at time zero of the time matrix with
        sub-steps.
        """
        assert all(time_mat_sub[0, 1:]==np.array([1, 0]*nfreq)[:-1])

    def test_behavior(self,):
        """Test that when the time matrix multiplies a state-vector it
        results in the correct response time-series.
        """
        f = 0.1
        w = 2*np.pi*f
        time_mat = wot.time_mat(f, 1)
        x = 1.2 + 3.4j
        X = np.reshape([0, np.real(x), np.imag(x)], [-1,1])[:-1]
        x_t = time_mat @ X
        t = wot.time(f, 1)
        assert np.allclose(x_t.squeeze(), np.real(x*np.exp(1j*w*t)))


class TestDerivativeMats:
    """Test functions :python:`derivative_mat`
    and :python:`derivative2_mat`.
    """

    @pytest.fixture(scope="class")
    def f1_dm(self,):
        """Fundamental frequency [Hz] for the synthetic derivative
        matrix.
        """
        return 0.5

    @pytest.fixture(scope="class")
    def nfreq_dm(self,):
        """Number of frequencies (harmonics) for the synthetic
        derivative matrix.
        """
        return 2

    @pytest.fixture(scope="class")
    def derivative_mat(self, f1_dm):
        """Correct/expected derivative matrix."""
        w0 = 2*np.pi*f1_dm
        mat = np.array([
            [0,  0,   0,    0],
            [0,  0, -w0,    0],
            [0, w0,   0,    0],
            [0,  0,   0,    0],
        ])
        return mat
    
    @pytest.fixture(scope="class")
    def derivative2_mat(self, f1_dm):
        """Correct/expected second derivative matrix."""
        w0 = 2*np.pi*f1_dm
        mat = np.array([
            [0,      0,      0,          0],
            [0, -w0**2,      0,          0],
            [0,      0, -w0**2,          0],
            [0,      0,      0, -(2*w0)**2],
        ])
        return mat

    def test_derivative_mat(self, derivative_mat, f1_dm, nfreq_dm):
        """Test the default created derivative matrix."""
        calculated = wot.derivative_mat(f1_dm, nfreq_dm)
        assert calculated==approx(derivative_mat)

    def test_derivative2_mat(self, derivative2_mat, f1_dm, nfreq_dm):
        """Test the default created second derivative matrix."""
        calculated = wot.derivative2_mat(f1_dm, nfreq_dm)
        assert calculated==approx(derivative2_mat)

    def test_dmat_no_mean(self, derivative_mat, f1_dm, nfreq_dm):
        """Test the derivative matrix without the mean component."""
        calculated = wot.derivative_mat(f1_dm, nfreq_dm, False)
        assert calculated==approx(derivative_mat[1:, 1:])

    def test_d2mat_no_mean(self, derivative2_mat, f1_dm, nfreq_dm):
        """Test the second derivative matrix without the mean component."""
        calculated = wot.derivative2_mat(f1_dm, nfreq_dm, False)
        assert calculated==approx(derivative2_mat[1:, 1:])

    def test_dmat_behavior(self,):
        """Test that when the derivative matrix multiplies a
        state-vector it results in the correct state-vector for the
        derivative of the input response.
        """
        f = 0.1
        w = 2*np.pi*f
        x = np.array([1 + 2j,
                      3 + 4j,
                      5 + 6j])
        derivative_mat = wot.derivative_mat(f, np.size(x))
        X = np.concatenate([
            [0.],
            np.reshape([[np.real(i), np.imag(i)] for i in x[:-1]], -1),
            [np.real(x[-1])]
        ])
        V = derivative_mat @ X
        v = np.sum(V[1::2]) + 1j*np.sum(V[2::2])
        expected = np.sum(
            [[(i+1) * 1j * w * x[i]] for i in range(np.size(x)-1)])
        assert np.allclose(v, expected)

    def test_d2mat_behavior(self,):
        """Test that when the second derivative matrix multiplies a
        state-vector it results in the correct state-vector for the
        second derivative of the input response.
        """
        f = 0.1
        w = 2*np.pi*f
        x = np.array([1 + 2j,
                        3 + 4j,
                        5 + 6j])
        derivative2_mat = wot.derivative2_mat(f, np.size(x))
        X = np.concatenate([
            [0.],
            np.reshape([[np.real(i), np.imag(i)] for i in x[:-1]], -1),
            [np.real(x[-1])]
        ])
        V = derivative2_mat @ X
        v = np.sum(V[1::2]) + 1j*np.sum(V[2::2])
        expected = np.sum(
            [[-(i * w)**2 * x[i]] for i in range(np.size(x)-1)]
            + np.real(np.size(x) * w)**2 * x[-1])


class TestMIMOTransferMat:
    """Test function :python:`mimo_transfer_mat`."""

    def test_mimo_transfer_mat(self, rao, mimo):
        """Test the function produces the correct MIMO transfer matrix.
        """
        calculated = wot.mimo_transfer_mat(rao, False)
        assert np.all(calculated == mimo)

    def test_behavior(self,):
        """Test that the MIMO transfer matrix applied to a state vector
        produces the expected output state vector, based on the
        synthetic RAO transfer function.
        """
        # test use/behavior
        x = np.array([1 + 2j,
                      3 + 4j])
        X = np.concatenate([
            [0.],
            np.reshape([[np.real(i), np.imag(i)] for i in x[:-1]], -1),
            [np.real(x[-1])]
        ])
        z = np.array([1.5+2.5j,
                      3.5+4.5j])
        F = np.concatenate([
            [0.],
            np.reshape(
                [[np.real(z[i]*x[i]), np.imag(z[i]*x[i])] for 
                i in range(np.size(x)-1)], -1),
            [np.real(z[-1]) * np.real(x[-1])],
        ])
        Z_mimo = wot.mimo_transfer_mat(np.reshape([z], [1,1,-1]), False)
        assert np.allclose(Z_mimo @ X, F)


class TestVecToDOFMatToVec:
    """Test functions :python:`vec_to_dofmat` and
    :python:`dofmat_to_vec`.
    """

    @pytest.fixture(scope="class")
    def vec(self,):
        """Sample vector."""
        return np.array([1,2,3,4,5,6])

    @pytest.fixture(scope="class")
    def dofmats(self, vec):
        """Correct sample matrices from the sample vector."""
        m1 = np.array([[1], [2], [3], [4], [5], [6]])
        m2 = np.array([
            [1, 4],
            [2, 5],
            [3, 6],
        ])
        m3 = np.array([
            [1, 3, 5],
            [2, 4, 6],
        ])
        return (m1, m2, m3)

    @pytest.fixture(scope="class")
    def dofmats_calc(self, vec):
        """Calculate sample matrices from the sample vector."""
        m1 = wot.vec_to_dofmat(vec, 1)
        m2 = wot.vec_to_dofmat(vec, 2)
        m3 = wot.vec_to_dofmat(vec, 3)
        return (m1, m2, m3)

    def test_shapes(self, dofmats_calc):
        """Test the shapes of the calculated matrices."""
        (m1, m2, m3) = dofmats_calc
        assert m1.shape==(6, 1) and m2.shape==(3, 2) and m3.shape==(2, 3)

    def test_values(self, dofmats, dofmats_calc):
        """Test the values of the calculated matrices."""
        (m1c, m2c, m3c) = dofmats_calc
        (m1, m2, m3) = dofmats
        assert np.all(m1c==m1) and np.all(m2c==m2) and np.all(m3c==m3)

    def test_recover_vec(self, vec, dofmats_calc):
        """Test that :python:`dofmat_to_vec` recovers the original
        vector.
        """
        (m1, m2, m3) = dofmats_calc
        v1 = wot.dofmat_to_vec(m1)
        v2 = wot.dofmat_to_vec(m2)
        v3 = wot.dofmat_to_vec(m3)
        assert np.all(v1==vec) and np.all(v2==vec) and np.all(v3==vec)

    def test_error(self, vec):
        """Test that the function raises a ValueError if an incompatible
        number of degrees of freedom are specified.
        """
        with pytest.raises(ValueError):
            wot.vec_to_dofmat(vec, 4)


class TestRealToComplexToReal:
    """Test functions :python:`real_to_complex` and
    :python:`complex_to_real`.
    """

    @pytest.fixture(scope="class")
    def complex_response(self,):
        """Sample complex response with 2 degrees of freedom and 2
        frequencies components plus the mean (zero-frequency) component.
        """
        response = np.array([
            # DOF 1,  DOF 2
            [  1+0j, 11+ 0j],  # f0
            [  2+3j, 12+13j],  # f1
            [  4   , 14    ],  # f2 (real component only)
        ])
        return response

    @pytest.fixture(scope="class")
    def real_response(self,):
        """Sample response as real coefficients (state vector used in
        WecOptTool).
        """
        response = np.array([
            # D1, D2
            [  1, 11],  # f0
            [  2, 12],  # f1 real
            [  3, 13],  # f1 imag
            [  4, 14],  # f2 real
        ])
        return response

    def test_complex_to_real(self, complex_response, real_response):
        """Test converting from complex to real."""
        calculated = wot.complex_to_real(complex_response)
        assert np.allclose(calculated, real_response)

    def test_real_to_complex(self, complex_response, real_response):
        """Test converting from real to complex."""
        calculated = wot.real_to_complex(real_response)
        assert np.allclose(calculated, complex_response)

    def test_cycle_real(self, real_response):
        """Test converting from real to complex and back to real."""
        calculated = wot.real_to_complex(real_response)
        calculated = wot.complex_to_real(calculated)
        assert np.allclose(calculated, real_response)

    def test_cycle_complex(self, complex_response):
        """Test converting from complex to real and back to complex."""
        calculated = wot.complex_to_real(complex_response)
        calculated = wot.real_to_complex(calculated)
        assert np.allclose(calculated, complex_response)

    def test_shapes(self, complex_response, real_response):
        """Test output shapes."""
        c_calc = wot.real_to_complex(real_response)
        r_calc = wot.complex_to_real(c_calc)
        c_shape = complex_response.shape
        r_shape = real_response.shape
        assert (c_calc.shape==c_shape) and (r_calc.shape==r_shape)

    def test_1d(self,):
        """Test the function with a 1-dimensional complex array as
        input.
        """
        complex_1d = np.array([1+0j, 2+3j])
        real_1d_calculated = wot.complex_to_real(complex_1d)
        real_1d = np.array([[1], [2]])
        assert np.allclose(real_1d_calculated, real_1d)


class TestFDToTDToFD:
    """Test functions :python:`fd_to_td` and :python:`td_to_fd`."""

    @pytest.fixture(scope="class")
    def components(self):
        """Values of the two non-zero components of the response."""
        a0r, a0i = 1, 2
        a1r, a1i = 3, 4
        return (a0r, a0i, a1r, a1i)

    @pytest.fixture(scope="class")
    def idx(self, nfreq):
        return np.random.randint(1, nfreq-1)
    
    @pytest.fixture(scope="class")
    def dc(self):
        return np.random.randint(1, 5)

    @pytest.fixture(scope="class")
    def fd(self, nfreq, components, idx):
        """Sample frequency domain response. There are two degrees of
        freedom and the response is zero at all but one (random)
        frequency. The non-zero values are set based on
        :python:`components`.
        """
        (a0r, a0i, a1r, a1i) = components
        fd = np.zeros([nfreq+1, 2], dtype=complex)
        fd[idx, 0] = a0r + a0i*1j
        fd[idx, 1] = a1r + a1i*1j
        return fd

    @pytest.fixture(scope="class")
    def td(self, nfreq, f1, components, idx):
        """Corresponding sample time domain response."""
        freq = wot.frequency(f1, nfreq)
        time = wot.time(f1, nfreq)
        wt  = 2*np.pi*freq[idx]*time
        (a0r, a0i, a1r, a1i) = components
        td = np.zeros([len(time), 2])
        td[:, 0] = a0r*np.cos(wt) - a0i*np.sin(wt)
        td[:, 1] = a1r*np.cos(wt) - a1i*np.sin(wt)
        return td

    @pytest.fixture(scope="class")
    def fd_1dof(self, nfreq, components, idx):
        """Sample frequency domain response with only one degree of
        freedom.
        """
        (a0r, a0i, _, _) = components
        fd = np.zeros(nfreq+1, dtype=complex)
        fd[idx] = a0r + a0i*1j
        return fd

    @pytest.fixture(scope="class")
    def td_1dof(self, nfreq, f1, components, idx):
        """Corresponding sample time domain response with only one
        degree of freedom.
        """
        (a0r, a0i, _, _) = components
        freq = wot.frequency(f1, nfreq)
        time = wot.time(f1, nfreq)
        wt  = 2*np.pi*freq[idx]*time
        return a0r*np.cos(wt) - a0i*np.sin(wt)

    @pytest.fixture(scope="class")
    def fd_topfreq(self, nfreq, components):
        """Sample frequency domain response with the nonzero components
        in the highest (Nyquist) frequency."""
        (a0r, _, a1r, _) = components
        fd = np.zeros([nfreq+1, 2], dtype=complex)
        fd[-1, 0] = a0r + 0j
        fd[-1, 1] = a1r + 0j
        return fd
    
    @pytest.fixture(scope="class")
    def td_topfreq(self, nfreq, f1, components):
        """Corresponding sample time domain response for the frequency
        vector with a nonzero top (Nyquist) frequency."""
        freq = wot.frequency(f1, nfreq)
        time = wot.time(f1, nfreq)
        wt  = 2*np.pi*freq[-1]*time
        (a0r, a0i, a1r, a1i) = components
        td = np.zeros([len(time), 2])
        td[:, 0] = a0r*np.cos(wt) - a0i*np.sin(wt)
        td[:, 1] = a1r*np.cos(wt) - a1i*np.sin(wt)
        return td
    
    @pytest.fixture(scope="class")
    def fd_nzmean(self, nfreq, components, idx, dc):
        """Sample frequency domain response with a nonzero mean."""
        (a0r, a0i, a1r, a1i) = components
        fd = np.zeros([nfreq+1, 2], dtype=complex)
        fd[0, :] = dc
        fd[idx, 0] = a0r + a0i*1j
        fd[idx, 1] = a1r + a1i*1j
        return fd

    @pytest.fixture(scope="class")
    def td_nzmean(self, nfreq, f1, components, idx, dc):
        """Corresponding sample time domain response with a nonzero mean."""
        freq = wot.frequency(f1, nfreq)
        time = wot.time(f1, nfreq)
        wt  = 2*np.pi*freq[idx]*time
        (a0r, a0i, a1r, a1i) = components
        td = np.zeros([len(time), 2])
        td[:, 0] = a0r*np.cos(wt) - a0i*np.sin(wt) + dc
        td[:, 1] = a1r*np.cos(wt) - a1i*np.sin(wt) + dc
        return td

    def test_fd_to_td(self, fd, td, f1, nfreq):
        """Test the :python:`fd_to_td` function outputs."""
        calculated = wot.fd_to_td(fd, f1, nfreq)
        assert calculated.shape==(2*nfreq, 2) and np.allclose(calculated, td)

    def test_td_to_fd(self, fd, td, nfreq):
        """Test the :python:`td_to_fd` function outputs."""
        calculated = wot.td_to_fd(td)
        assert calculated.shape==(nfreq+1, 2) and np.allclose(calculated, fd)

    def test_fft(self, fd, td, nfreq):
        """Test the :python:`fd_to_td` function outputs when using FFT.
        """
        calculated = wot.fd_to_td(fd)
        assert calculated.shape==(2*nfreq, 2) and np.allclose(calculated, td)

    def test_fd_to_td_1dof(self, fd_1dof, td_1dof, f1, nfreq):
        """Test the :python:`fd_to_td` function outputs for the 1 DOF
        case.
        """
        calculated = wot.fd_to_td(fd_1dof, f1, nfreq)
        shape = (2*nfreq, 1)
        calc_flat = calculated.squeeze()
        assert calculated.shape==shape and np.allclose(calc_flat, td_1dof)

    def test_td_to_fd_1dof(self, fd_1dof, td_1dof, nfreq):
        """Test the :python:`td_to_fd` function outputs for the 1 DOF
        case.
        """
        calculated = wot.td_to_fd(td_1dof.squeeze())
        shape = (nfreq+1, 1)
        calc_flat = calculated.squeeze()
        assert calculated.shape==shape and np.allclose(calc_flat, fd_1dof)

    def test_fft_1dof(self, fd_1dof, td_1dof, nfreq):
        """Test the :python:`fd_to_td` function outputs when using FFT
        for the 1 DOF.
        """
        calculated = wot.fd_to_td(fd_1dof)
        shape = (2*nfreq, 1)
        calc_flat = calculated.squeeze()
        assert calculated.shape==shape and np.allclose(calc_flat, td_1dof)

    def test_fd_to_td_nzmean(self, fd_nzmean, td_nzmean, f1, nfreq):
        """Test the :python: `td_to_fd` function outputs with a 
        nonzero mean value.
        """
        calculated = wot.fd_to_td(fd_nzmean, f1, nfreq)
        assert calculated.shape==(2*nfreq, 2) and np.allclose(calculated, td_nzmean)

    def test_td_to_fd_nzmean(self, fd_nzmean, td_nzmean, nfreq):
        """Test the :python: `td_to_fd` function outputs with a
        nonzero mean value.
        """
        calculated = wot.td_to_fd(td_nzmean)
        assert calculated.shape==(nfreq+1, 2) and np.allclose(calculated, fd_nzmean)

    def test_fd_to_td_nzmean(self, fd_nzmean, td_nzmean, f1, nfreq):
        """Test the :python: `td_to_fd` function outputs with the top (Nyquist)
        frequency vector.
        """
        calculated = wot.fd_to_td(fd_nzmean, f1, nfreq)
        assert calculated.shape==(2*nfreq, 2) and np.allclose(calculated, td_nzmean)

    def test_td_to_fd_topfreq(self, fd_topfreq, td_topfreq, nfreq):
        """Test the :python: `td_to_fd` function outputs for the
        Nyquist frequency.
        """
        calculated = wot.td_to_fd(td_topfreq)
        assert calculated.shape==(nfreq+1, 2) and np.allclose(calculated, fd_topfreq)


class TestReadWriteNetCDF:
    """Test functions :python:`read_netcdf` and :python:`write_netcdf`.
    """

    @pytest.fixture(scope="class")
    def hydro_data_new(self, hydro_data):
        """Hydrodynamic data structure created by writing and then
        reading data to a NetCDF file.
        """
        cwd = os.path.dirname(__file__)
        bem_file_tmp = os.path.join(cwd, 'bem_tmp.nc')
        wot.write_netcdf(bem_file_tmp, hydro_data)
        hydro_data_new = wot.read_netcdf(bem_file_tmp)
        os.remove(bem_file_tmp)
        return hydro_data_new

    def test_read_write_netcdf(self, hydro_data, hydro_data_new):
        """Test that writing to, followed by reading from, a NetCDF file
        results in the same original hydrodynamic data structure.
        """
        assert hydro_data.equals(hydro_data_new)


class TestCheckLinearDamping:
    """Test function :python:`check_linear_damping`."""

    @pytest.fixture(scope="class")
    def data(self, hydro_data):
        """Hydrodynamic data structure for this test set."""
        data = hydro_data.copy(deep=True)
        data['radiation_damping'] *= 0
        data['friction'] *= 0
        data['friction'] += -0.1
        return data

    @pytest.fixture(scope="class")
    def tol(self, data):
        """Tolerance for function :python:`check_linear_damping`."""
        return 0.01

    @pytest.fixture(scope="class")
    def data_new_uniform(self, data, tol):
        """Hydrodynamic data structure for which the function
        :python:`check_linear_damping` has been called.
        """
        return wot.check_linear_damping(data, tol)

    @pytest.fixture(scope="class")
    def data_new_nonuniform(self, data, tol):
        """Hydrodynamic data structure for which the function
        :python:`check_linear_damping` has been called.
        """
        # TODO: clean this up when fixing the dim order discrepancy
        data['radiation_damping'] = data['radiation_damping'].transpose('omega', ...)
        return wot.check_linear_damping(data, tol, False)

    def test_friction(self, data_new_uniform, tol):
        """Test that the modified friction diagonal has the expected
        value for a uniform shift.
        """
        assert np.allclose(np.diagonal(data_new_uniform.friction.values), tol)

    def test_only_diagonal_friction(self, data, data_new_uniform):
        """Test that only the diagonal was changed for a uniform shift."""
        data_org = data.copy(deep=True)
        def nodiag(x):
            return x.friction.values - np.diag(np.diagonal(x.friction.values))
        assert np.allclose(nodiag(data_new_uniform), nodiag(data_org))

    def test_only_friction(self, data, data_new_uniform):
        """Test that only the friction is changed in the hydrodynamic
        data for a uniform shift.
        """
        data_new_nofric = data_new_uniform.copy(deep=True
                                                ).drop_vars('friction')
        data_org_nofric = data.copy(deep=True).drop_vars('friction')
        assert data_new_nofric.equals(data_org_nofric)

    def test_damping(self, data_new_nonuniform, tol):
        """Test that the modified radiation damping diagonal has the expected
        value for a non-uniform shift.
        """
        assert np.allclose(
            np.diagonal(data_new_nonuniform.radiation_damping.values,
                        axis1=1, axis2=2),
                        tol)

    def test_only_diagonal_damping(self, data_new_nonuniform):
        """Test that no off-diagonal radiation damping terms are nonzero
        for a non-uniform shift.
        """
        assert (
            np.prod(
            np.shape(data_new_nonuniform.radiation_damping.values)[:-1]) == 
            np.count_nonzero(data_new_nonuniform.radiation_damping.values)
        )

    def test_only_rd(self, data, data_new_nonuniform):
        """Test that only the radiation damping is changed in the hydrodynamic
        data for a non-uniform shift.
        """
        data_new_nord = data_new_nonuniform.copy(
            deep=True).drop_vars('radiation_damping')
        data_org_nord = data.copy(deep=True).drop_vars('radiation_damping')
        assert data_new_nord.equals(data_org_nord)
        

class TestCheckImpedance:
    """Test functions :python:`hydrodynamic_impedance` and 
    :python:`check_impedance`."""

    @pytest.fixture(scope="class")
    def data(self, hydro_data):
        """Hydrodynamic data structure for this test set."""
        data = hydro_data.copy(deep=True)
        data['radiation_damping'] *= 0
        data['friction'] *= 0
        data['friction'] += -0.1
        Zi = wot.hydrodynamic_impedance(data)
        return Zi
    
    def test_hydrodynamic_impedance(self, data, hydro_data):
        """"Check that shape of impedance is as expected"""
        assert data.shape == hydro_data.added_mass.shape

    @pytest.fixture(scope="class")
    def tol(self, data):
        """Tolerance for function :python:`check_impedance`."""
        return 0.01

    @pytest.fixture(scope="class")
    def data_new(self, data, tol):
        """Hydrodynamic data structure for which the function
        :python:`check_impedance` has been called.
        """
        return wot.check_impedance(data, tol)

    def test_friction(self, data_new, tol):
        """Test that the modified impedance diagonal has the expected
        value.
        """
        assert np.allclose(np.real(np.diagonal(data_new)), tol)

    def test_only_diagonal_friction(self, data, data_new):
        """Test that only the diagonal was changed."""
        data_org = data.copy(deep=True)

        def offdiags(x):
            return x.values[np.invert(np.eye(x.shape[0], dtype=bool))]
        assert np.allclose(offdiags(data_new), offdiags(data_org))

    def test_only_friction(self, data, data_new):
        """Test that only the real part of the impedance was changed.
        """
        assert np.allclose(np.imag(data), np.imag(data_new))


class TestForceFromImpedanceOrTransferFunction:
    """Test both functions :python:`force_from_rao_transfer_function`
    and :python:`force_from_impedance`.
    """

    @pytest.fixture(scope="class")
    def omega(self, f1, nfreq_imp):
        """Radial frequency vector."""
        return wot.frequency(f1, nfreq_imp) * 2*np.pi

    @pytest.fixture(scope="class")
    def x_wec(self,):
        """WEC position state vector for a simple synthetic case."""
        return [0, 1, 1, 1, 0, 2, 2, 2]

    @pytest.fixture(scope="class")
    def force(self, f1, nfreq_imp, ndof_imp):
        """Correct forces for the synthetic example."""
        # amplitude: A  = mimo @ x_wec, calculated manually
        #   reshaped for convenience
        A = np.array([
            [0, 1,  7, 44],
            [0, 4, 10, 77],
        ])
        force = np.zeros((wot.ncomponents(nfreq_imp), ndof_imp))
        w = wot.frequency(f1, nfreq_imp) * 2*np.pi
        t = wot.time(f1, nfreq_imp)
        force[:, 0] = (
            A[0, 0] +
            A[0, 1]*np.cos(w[1]*t) - A[0, 2]*np.sin(w[1]*t) +
            A[0, 3]*np.cos(w[2]*t)
        )
        force[:, 1] = (
            A[1, 0] +
            A[1, 1]*np.cos(w[1]*t) - A[1, 2]*np.sin(w[1]*t) +
            A[1, 3]*np.cos(w[2]*t)
        )
        return force

    def test_from_transfer(
            self, rao, f1, nfreq_imp, ndof_imp, force, x_wec
        ):
        """Test the function :python:`force_from_rao_transfer_function`
        for a small synthetic problem.
        """
        force_func = wot.force_from_rao_transfer_function(rao, False)
        wec = wot.WEC(f1, nfreq_imp, {}, ndof=ndof_imp, inertia_in_forces=True)
        force_calculated = force_func(wec, x_wec, None, None)
        assert np.allclose(force_calculated, force)

    def test_from_impedance(
            self, rao, f1, nfreq_imp, ndof_imp, force, x_wec, omega
        ):
        """Test the function :python:`force_from_impedance` for a small
        synthetic problem.
        """
        force_func = wot.force_from_impedance(omega[1:], rao/(1j*omega[1:]))
        wec = wot.WEC(f1, nfreq_imp, {}, ndof=ndof_imp, inertia_in_forces=True)
        force_calculated = force_func(wec, x_wec, None, None)
        assert np.allclose(force_calculated, force)


class TestForceFromWaves:
    """Test function :python:`force_from_waves`."""

    def test_regular(
            self, exc_coeff, f1, nfreq, ndof_waves, wave_regular, fexc_regular
        ):
        """Test regular wave forces."""
        # correct
        A = fexc_regular
        force = np.zeros((wot.ncomponents(nfreq), ndof_waves))
        w = wot.frequency(f1, nfreq) * 2*np.pi
        w = w[1:]
        t = wot.time(f1, nfreq)
        for i in range(ndof_waves):
            for j in range(nfreq):
                Ar, Ai = np.real(A[j, i]), np.imag(A[j, i])
                force[:, i] += Ar*np.cos(w[j]*t) - Ai*np.sin(w[j]*t)
        # calculated
        force_func = wot.force_from_waves(exc_coeff)
        wec = wot.WEC(f1, nfreq, {}, ndof=ndof_waves, inertia_in_forces=True)
        waves, _ = wave_regular
        force_calculated = force_func(wec, None, None, waves)
        # test
        assert np.allclose(force_calculated, force)

    def test_multi(
            self, exc_coeff, f1, nfreq, ndof_waves, waves_multi, fexc_multi
        ):
        """Test iregular wave forces."""
        # correct
        A = fexc_multi
        force = np.zeros((wot.ncomponents(nfreq), ndof_waves))
        w = wot.frequency(f1, nfreq) * 2*np.pi
        w = w[1:]
        t = wot.time(f1, nfreq)
        for i in range(ndof_waves):
            for j in range(nfreq):
                Ar, Ai = np.real(A[j, i]), np.imag(A[j, i])
                force[:, i] += Ar*np.cos(w[j]*t) - Ai*np.sin(w[j]*t)
        # calculated
        force_func = wot.force_from_waves(exc_coeff)
        wec = wot.WEC(f1, nfreq, {}, ndof=ndof_waves, inertia_in_forces=True)
        waves, _ = waves_multi
        force_calculated = force_func(wec, None, None, waves)
        # test
        assert np.allclose(force_calculated, force)


class TestInertiaStandardForces:
    """Test functions :python:`inertia` and :python:`standard_forces`.
    """

    @pytest.fixture(scope="class")
    def forces(self, hydro_data):
        data = hydro_data.copy(deep=True)
        nfreq = len(data.omega)
        f1 = data.omega.values[0] / (2*np.pi)
        ndof = len(data.influenced_dof)
        ndir = len(data.wave_direction)
        # time series of responses
        index_freq = np.random.randint(1, nfreq)
        amplitude = 1.3
        w = data.omega.values[index_freq-1]
        t = wot.time(f1, nfreq)
        pos = amplitude * np.cos(w*t)
        vel = -1*w*amplitude * np.sin(w*t)
        acc = -1*w**2*amplitude * np.cos(w*t)
        # parameters
        mass = 2.1
        rad = 1.2
        addmass = 0.7
        hstiff = 0.9
        fric = 3.4
        wave_freq = w/(2*np.pi)
        wave_amp = 1.1
        wave_phase = 0.46
        wave_phase_deg = np.rad2deg(wave_phase)
        diff = 4 + 5j
        fk_coeff = -2 + 1.2j
        # build hydrodynamic dataset
        data['inertia_matrix'][:, :] = np.eye(ndof)*mass
        data['hydrostatic_stiffness'][:, :] = np.eye(ndof)*hstiff
        data['friction'][:, :] = np.eye(ndof)*fric
        data['radiation_damping'].values[:, :, index_freq-1] = np.eye(ndof)*rad
        data['added_mass'].values[:, :, index_freq-1] = np.eye(ndof)*addmass
        data['diffraction_force'].values[:, :, index_freq-1] = (
            np.zeros([ndof, ndir], dtype=complex))
        data['diffraction_force'].values[0, 0, index_freq-1] = diff
        data['Froude_Krylov_force'].values[:, :, index_freq-1] = (
            np.zeros([ndof, ndir], dtype=complex))
        data['Froude_Krylov_force'].values[0, 0, index_freq-1] = fk_coeff
        # standard forces
        forces = wot.standard_forces(data)
        # calculated inertia and forces
        wec = wot.WEC(
            f1, nfreq, {}, inertia_matrix=data['inertia_matrix'].values)
        waves = wot.waves.regular_wave(
            f1, nfreq, wave_freq, wave_amp, wave_phase_deg)
        x_wec = np.zeros((nfreq*2)*ndof)
        x_wec[(index_freq*2-1)*1] = amplitude
        inertia_func = wot.inertia(
            f1, nfreq, inertia_matrix=data['inertia_matrix'].values)
        inertia = inertia_func(wec, x_wec, None, None)
        radiation = forces['radiation'](wec, x_wec, None, None)
        hydrostatics = forces['hydrostatics'](wec, x_wec, None, None)
        friction = forces['friction'](wec, x_wec, None, None)
        fk = forces['Froude_Krylov'](wec, None, None, waves)
        diffraction = forces['diffraction'](wec, None, None, waves)
        # true/expected inertia and forces
        inertia_truth = mass * acc
        radiation_truth = -(rad*vel + addmass*acc)
        hydrostatics_truth = -hstiff*pos
        friction_truth = -fric*vel
        diff_comp = wave_amp*np.exp(1j*wave_phase) * diff
        diffraction_truth =  (np.real(diff_comp) * np.cos(w*t) -
                                np.imag(diff_comp) * np.sin(w*t))
        fk_comp = wave_amp*np.exp(1j*wave_phase) * fk_coeff
        fk_truth =  np.real(fk_comp) * np.cos(w*t) - np.imag(fk_comp) * np.sin(w*t)
        # output dictionary
        forces = [
            (inertia, inertia_truth, "inertia"),
            (radiation, radiation_truth, "radiation"),
            (hydrostatics, hydrostatics_truth, "hydrostatics"),
            (friction, friction_truth, "friction"),
            (diffraction, diffraction_truth, "diffraction"),
            (fk, fk_truth, "")
        ]
        return forces

    def test_values(self, forces):
        """Test the values of inertia and all standard forces."""
        for f_calc, f_truth, name in forces:
            assert np.allclose(f_calc[:, 0], f_truth), f"FAILED: {name}!"

    def test_mean(self, forces):
        """Test that the mean component of inertia and all standard
        forces is zero.
        """
        for f_calc, _, name in forces:
            assert np.allclose(f_calc[:, 1], 0), f"FAILED: {name}!"


class TestRunBEM:
    """Test function :python:`run_bem`."""

    def test_it_runs(self,):
        """Test that the function at least runs and returns correct
        data type.
        """
        rect = cpy.RectangularParallelepiped(
            size=(5.0, 5.0, 2.0), resolution=(10, 10, 10), center=(0.0, 0.0, 0.0,)
        )
        rect.add_translation_dof(name="Heave")
        bem_data = wot.run_bem(fb=rect, freq=[0.1, 0.2], wave_dirs=[0,])
        assert type(bem_data) == xr.Dataset


class TestChangeBEMConvention:
    """Test function :python:`change_bem_convention`."""

    @pytest.fixture(scope="class")
    def data_there(self, bem_data):
        """Modified BEM data."""
        data_here = bem_data.copy(deep=True)
        return wot.change_bem_convention(data_here)

    @pytest.fixture(scope="class")
    def data_back(self, data_there):
        """Twice modified BEM data. Should be equal to original."""
        return wot.change_bem_convention(data_there)

    def test_fk(self, bem_data, data_there):
        """Test that the Froude-Krylov force was changed correctly."""
        calculated = data_there['Froude_Krylov_force'].values
        correct = np.conjugate(bem_data['Froude_Krylov_force'].values)
        assert np.allclose(calculated, correct)

    def test_diff(self, bem_data, data_there):
        """Test that the diffraction force was changed correctly."""
        calculated = data_there['diffraction_force'].values
        correct = np.conjugate(bem_data['diffraction_force'].values)
        assert np.allclose(calculated, correct)

    def test_round_trip(self, bem_data, data_back):
        """Test that a round trip returns the original dataset."""
        xr.testing.assert_allclose(data_back, bem_data)


class TestLinearHydrodynamics:
    """Test function :python:`linear_hydrodynamics`."""

    def test_values(self, bem_data, hydro_data):
        """Test the function returns expected values."""
        mat = np.array([[1, 1], [1, 1]])
        calculated = wot.linear_hydrodynamics(bem_data, mat, mat, mat)
        xr.testing.assert_allclose(calculated, hydro_data)


class TestWaveExcitation:
    """Test function :python:`wave_excitation`."""

    def test_regular_value(
            self, exc_coeff, wave_regular, fexc_regular, nfreq, ndof_waves
        ):
        """Test the value of the regular wave excitation force."""
        wave_regular, _ = wave_regular
        calculated = wot.wave_excitation(exc_coeff, wave_regular)
        calc_shape = calculated.shape
        shape = (nfreq, ndof_waves)
        assert calc_shape==shape and np.allclose(calculated, fexc_regular)

    def test_multi_value(
            self, exc_coeff, waves_multi, fexc_multi, nfreq, ndof_waves
        ):
        """Test the value of the multi-directional wave excitation
        force.
        """
        waves_multi, _ = waves_multi
        calculated = wot.wave_excitation(exc_coeff, waves_multi)
        calc_shape = calculated.shape
        shape = (nfreq, ndof_waves)
        assert calc_shape==shape and np.allclose(calculated, fexc_multi)

    def test_error_directions_not_subset(self, f1, nfreq, exc_coeff):
        """Test that an error is raised if the wave directions are not
        a subset of the hydrodata directions.
        """
        with pytest.raises(ValueError):
            waves = wot.waves.elevation_fd(f1, nfreq, [0.0, 25])
            wot.wave_excitation(exc_coeff, waves)

    def test_error_different_frequencies(
            self, f1, nfreq, waves_multi, exc_coeff
        ):
        """Test that an error is raised if the wave and hydrodata do not
        have the same frequencies.
        """
        _, params_multi = waves_multi
        directions = np.deg2rad(params_multi['directions'])
        with pytest.raises(ValueError):
            waves = wot.waves.elevation_fd(f1+0.01, nfreq, directions)
            wot.wave_excitation(exc_coeff, waves)


class TestAtleast2D:
    """Test function :python:`atleast_2d`."""

    def test_1d(self,):
        """Test function creates extra trailing dimension when a 1D array
        is passed.
        """
        a = np.array([1.1, 2.2, 3.3])
        b = wot.atleast_2d(a)
        assert a.shape!=b.shape and b.shape==(len(a), 1)

    def test_2d(self,):
        """Test function is identity when a 2D array is passed.
        """
        a = np.array([[1.1, 2.2, 3.3]])
        b = wot.atleast_2d(a)
        assert b.shape==a.shape and np.allclose(b, a)


class TestDegreesToRadians:
    """Test function :python:`degrees_to_radians`."""

    @pytest.fixture(scope="class")
    def degree(self,):
        """Single angle in degrees."""
        return 44.26

    @pytest.fixture(scope="class")
    def degrees(self,):
        """List of several angles in degrees."""
        return [0.0, -1000, 2.53, 9.99*np.pi]

    @pytest.fixture(scope="class")
    def rad(self, degree):
        """Single angle in radians."""
        return wot.degrees_to_radians(degree)

    @pytest.fixture(scope="class")
    def rads(self, degrees):
        """List of several angles in radians."""
        return wot.degrees_to_radians(degrees)


    def test_default_sort(self, degrees, rads):
        """Test default sorting behavior."""
        rads_sorted = wot.degrees_to_radians(degrees, sort=True)
        assert rads_sorted==approx(rads)

    def test_unsorted(self, degrees, rads):
        """Test unsorted behavior."""
        rads_us = wot.degrees_to_radians(degrees, sort=False)
        assert rads_us!=approx(rads) and np.sort(rads_us)==approx(rads)

    def test_numpy_array(self, degrees, rads):
        """Test function with a NumPy array."""
        rads_np = wot.degrees_to_radians(np.array(degrees))
        assert rads_np==approx(rads)

    def test_scalar(self, degree, rad):
        """Test function with a scalar."""
        rads_s = wot.degrees_to_radians(degree)
        assert rads_s==approx(rad)

    def test_special_cases(self,):
        """Test edge angles."""
        for r,d in [(0, 0), (np.pi, 180), (0, 360), (np.pi, -180)]:
            assert r==approx(wot.degrees_to_radians(d)), f"Failed: ({r}, {d})"

    def test_cyclic(self, degree, rad):
        """Test that cyclic permutations give same answer."""
        rad_cyc = wot.degrees_to_radians(degree+random.randint(-10,10)*360)
        assert rad_cyc==approx(rad)

    def test_range(self, rads):
        """Test that the outputs are in the range [-π, π) radians."""
        assert (all(rads>=-np.pi) and all(rads<np.pi))


class TestSubsetClose:
    """Test function :python:`subset_close`."""

    @pytest.fixture(scope="class")
    def array_a(self,):
        """An array of numbers."""
        return np.array([1, 2.35, 0.3])

    @pytest.fixture(scope="class")
    def array_b(self, array_a):
        """A second array that is longer and contains all elements of
        the first array.
        """
        n = 10 - len(array_a)
        b = np.concatenate([array_a, np.random.random(n)])
        np.random.shuffle(b)
        return b

    def test_subset(self, array_a, array_b):
        """Test subset identified correctly."""
        subset, _ = wot.subset_close(array_a, array_b)
        assert subset

    def test_not_subset(self, array_a, array_b):
        """Test `False` when it is not a subset."""
        subset, _ = wot.subset_close(array_b, array_a)
        assert not subset

    def test_atol(self, array_a, array_b):
        """Test the tolerance is used correctly."""
        tol = 0.001
        subset_t, _ = wot.subset_close(array_a, array_b+0.9*tol, atol=tol)
        subset_f, _ = wot.subset_close(array_a, array_b+1.1*tol, atol=tol)
        assert subset_t and not subset_f

    def test_indices(self,):
        """Test the correct indices are returned."""
        a = np.array([1.1, 2.2, 3.3])
        b = np.arange(10, dtype=float)
        idx = [1, 4, 6]
        for i,id in enumerate(idx):
            b[id] = a[i]
        subset, idx_calculated = wot.subset_close(a, b)
        assert subset and idx_calculated==idx

    def test_scalar(self,):
        """Test function when array_a is a scalar."""
        a = 2.2
        b = np.array([1.1, a, 3.3])
        subset, _ = wot.subset_close(a, b)
        subset_a, _ = wot.subset_close(a, a)
        assert subset and subset_a


class TestScaleDOFs:
    """Test function :python:`scale_dofs`."""

    def test_function(self,):
        """Test that the function returns expected results."""
        scales = [1.1, 2.2, 3.3]
        ncomponents = 3
        scale_vec_calculated = wot.scale_dofs(scales, ncomponents)
        scale_vec = [1.1, 1.1, 1.1, 2.2, 2.2, 2.2, 3.3, 3.3, 3.3]
        assert np.allclose(scale_vec_calculated, scale_vec)


class TestDecomposeState:
    """Test function :python:`decompose_state`."""

    def test_function(self,):
        """Test that the function returns expected results."""
        ndof, nfreq = 1, 2  # ncomponents = ndof*(2*nfreq-1)+1 = 4
        state = [1, 1, 1, 1, 3.4]
        x_wec = [1, 1, 1, 1]
        x_opt = [3.4]
        x_w_calc, x_o_calc = wot.decompose_state(state, ndof, nfreq)
        assert np.allclose(x_w_calc, x_wec) and np.allclose(x_o_calc, x_opt)


class TestFrequencyParameters:
    """Test function :python:`frequency_parameters`."""

    def test_default(self,):
        """Test that the function returns the same parameters used to
        create the frequency array.
        """
        f1 = np.random.random()
        nfreq = np.random.randint(2,10)
        freq = wot.frequency(f1, nfreq)
        f1_calc, nfreq_calc = wot.frequency_parameters(freq)
        assert f1_calc==approx(f1) and nfreq_calc==nfreq

    def test_zero_freq(self,):
        """Test with a frequency array not containing the zero frequency.
        """
        f1 = 0.1
        nfreq = 3
        freq = [0.1, 0.2, 0.3]
        f1_calc, nfreq_calc = wot.frequency_parameters(freq, False)
        assert f1_calc==approx(f1) and nfreq_calc==nfreq

    def test_error_spacing(self,):
        """Test that it throws an error if the frequency array is not
        evenly-spaced.
        """
        with pytest.raises(ValueError):
            freq = [0, 0.1, 0.2, 0.4]
            wot.frequency_parameters(freq)

    def test_error_zero(self,):
        """Test that it throws an error if the frequency array does not
        contain the zero-frequency.
        """
        with pytest.raises(ValueError):
            freq = [0.1, 0.2, 0.3]
            wot.frequency_parameters(freq)

    def test_error_zero_false(self,):
        """Test that it throws an error if the frequency array contains
        the zero-frequency but :python:`zero_freq` is :python:`False`.
        """
        with pytest.raises(ValueError):
            freq = [0, 0.1, 0.2, 0.3]
            wot.frequency_parameters(freq, False)


class TestTimeResults:
    """Test function :python:`time_results`."""

    @pytest.fixture(scope="class")
    def f1(self,):
        """Fundamental frequency [Hz]."""
        return 0.1

    @pytest.fixture(scope="class")
    def nfreq(self,):
        """Number of frequencies."""
        return 2

    @pytest.fixture(scope="class")
    def time(self, f1, nfreq):
        """Time vector [s]."""
        time = wot.time(f1, nfreq)
        return xr.DataArray(data=time, name='time', dims='time', coords=[time])

    @pytest.fixture(scope="class")
    def components(self,):
        """Real and imaginary components of the complex response."""
        return {'re1': 1.3, 'im1': -2.1, 're2': 0.5, 'im2': 0.4}

    @pytest.fixture(scope="class")
    def fd(self, f1, nfreq, components):
        """Frequency domain response."""
        omega = wot.frequency(f1, nfreq) * 2*np.pi
        re1 = components['re1']
        im1 = components['im1']
        re2 = components['re2']
        im2 = components['im2']
        mag = np.array([0.0, re1+im1*1j, re2+im2*1j])
        mag = xr.DataArray(
            data=mag, name='response', dims='omega', coords=[omega])
        return mag

    def test_values(self, f1, nfreq, time, fd, components):
        """Test that the function returns the correct time domain
        response.
        """
        td = wot.time_results(fd, time)
        re1 = components['re1']
        im1 = components['im1']
        re2 = components['re2']
        im2 = components['im2']
        w = wot.frequency(f1, nfreq) * 2*np.pi
        t = td.time.values
        response = (
            re1*np.cos(w[1]*t) - im1*np.sin(w[1]*t) +
            re2*np.cos(w[2]*t) - im2*np.sin(w[2]*t)
        )
        assert np.allclose(td.values, response)



"""Functions for calculating hydrostatic and mass properties for
floating bodies.
"""


from __future__ import annotations


__all__ = [
    "stiffness_matrix",
    "inertia_matrix"
]


from typing import Iterable, Optional
import logging

import numpy as np
from capytaine import FloatingBody
from xarray import DataArray

from wecopttool.core import _default_parameters


# logger
_log = logging.getLogger(__name__)

def stiffness_matrix(
    fb: FloatingBody,
    rho: float = _default_parameters['rho'],
    g: float = _default_parameters['g'],
    center_of_mass: Optional[Iterable[float]] = None,
    rotation_center: Optional[Iterable[float]] = None
) -> DataArray:
    """Compute the hydrostatic stiffness of a Capytaine floating body.

    .. note:: Only works for rigid body DOFs which must be named
              according to the Capytaine convention (e.g.,
              :python:`"Heave"`).

    Uses
    :py:meth:`capytaine.bodies.bodies.FloatingBody.compute_hydrostatic_stiffness`
    on the immersed part of the mesh.

    Parameters
    ----------
    fb
        A capytaine floating body (mesh + DOFs, and optionally center of
        mass).
    rho
        Water density in :math:`kg/m^3`.
    g
        Gravitational acceleration in :math:`m/s^2`.
    center_of_mass
        Center of gravity/mass :python:`(cx, cy, cz)`.
    rotation_center
        Center of rotation :python:`(rx, ry, rz)` for hydrostatics
        calculations.

    Raises
    ------
    ValueError
        If :python:`fb.center_of_mass is not None` and
        :python:`center_of_mass` is provided with a different value.
    """
    fb = _set_property(fb, 'center_of_mass', center_of_mass)
    fb = _set_property(fb, 'rotation_center', rotation_center)
    fb_im = fb.copy(name=f"{fb.name}_immersed").keep_immersed_part()
    return fb_im.compute_hydrostatic_stiffness(rho=rho, g=g)


def inertia_matrix(
    fb: FloatingBody,
    rho: Optional[float] = _default_parameters['rho'],
    center_of_mass: Optional[Iterable[float]] = None,
    mass: Optional[float] = None,
    rotation_center: Optional[Iterable[float]] = None,
) -> DataArray:
    """Compute the inertia (mass) matrix assuming a constant density for
    the WEC.

    .. note:: This function assumes a constant density WEC.

    Uses
    :py:meth:`capytaine.bodies.bodies.FloatingBody.compute_rigid_body_inertia`
    on the full mesh.

    Parameters
    ----------
    fb
        A capytaine floating body (mesh + DOFs, and optionally center of
        mass and mass).
    rho
        Water density in :math:`kg/m^3`.
    center_of_mass
        Center of gravity/mass.
    mass
        Rigid body mass.
    rotation_center
        Center of rotation :python:`(rx, ry, rz)` for hydrostatics
        calculations.

    Raises
    ------
    ValueError
        If :python:`fb.center_of_mass is not None` and
        :python:`center_of_mass` is provided with a different value.
    ValueError
        If :python:`fb.mass is not None` and :python:`mass` is provided
        with a different value.
    """
    fb = _set_property(fb, 'center_of_mass', center_of_mass)
    fb = _set_property(fb, 'rotation_center', rotation_center)
    fb = _set_property(fb, 'mass', mass, rho)
    return fb.compute_rigid_body_inertia(rho=rho)


def _set_property(
    fb: FloatingBody,
    property: str,
    value: Optional[Iterable[float]],
    rho: float = _default_parameters["rho"],
) -> FloatingBody:
    """Sets default properties if not provided by the user:
        - `center_of_mass` is set to the geometric centroid
        - `mass` is set to the displaced mass
        - `rotation_center` is set to the center of mass
    """
    valid_properties = ['mass', 'center_of_mass', 'rotation_center']
    if property not in valid_properties:
        raise ValueError(
            "`property` is not a recognized property. Valid properties are " +
           f"{valid_properties}."
        )
    if not hasattr(fb, property):
        setattr(fb, property, None)
    prop_org = getattr(fb, property) is not None
    prop_new = value is not None

    if not prop_org and not prop_new:
        if property == 'mass':
            vol = fb.copy(
                name=f"{fb.name}_immersed"
                ).keep_immersed_part().volume
            def_val = rho * vol
            log_str = (
                "Setting the mass to the displaced mass.")
        elif property == 'center_of_mass':
            def_val = fb.center_of_buoyancy
            log_str = (
                "Using the geometric centroid as the center of gravity (COG).")
        elif property == 'rotation_center':
            def_val = fb.center_of_mass
            log_str = (
                "Using the center of gravity (COG) as the rotation center " +
                "for hydrostatics.")
        setattr(fb, property, def_val)
        _log.info(log_str)
    elif prop_org and prop_new:
        if not np.allclose(getattr(fb, property), value):
            raise ValueError(
               f"Both :python:`fb.{property}` and " +
               f":python:`{property}` were provided but have " +
                "different values."
            )
    elif prop_new:
        setattr(fb, property, value)

    return fb



""" Unit tests for functions in the :python:`waves.py` module.
"""

import os

import pytest
import capytaine as cpy
import numpy as np
import wavespectra as ws
from scipy import signal

import wecopttool as wot
from wecopttool.core import _default_parameters


# setup wave structure
@pytest.fixture(scope="module")
def f1():
    """Fundamental frequency [Hz]."""
    return 0.12


@pytest.fixture(scope="module")
def nfreq():
    """Number of frequencies in frequency vector."""
    return 5


@pytest.fixture(scope="module")
def fp():
    """Peak frequency [Hz] of the sea state."""
    return 0.25


@pytest.fixture(scope="module")
def hs():
    """Significan wave height [m] of the sea state."""
    return 0.1


@pytest.fixture(scope="module")
def ndir():
    """Number of directions to consider in the range [0, 360)
    degrees.
    """
    return 90


@pytest.fixture(scope="module")
def directions(ndir):
    """Vector of wave directions."""
    return np.linspace(0, 360, ndir, endpoint=False)


class TestElevationFD:
    """Test function :python:`waves.elevation_fd`."""

    @pytest.fixture(scope="class")
    def elevation(self, f1, nfreq, directions):
        """Complex sea state elevation amplitude [m] indexed by
        frequency and direction."""
        return wot.waves.elevation_fd(f1, nfreq, directions)

    def test_coordinates(self, elevation):
        """Test that the elevation dataArray has the correct
        coordinates.
        """
        coordinates = ['wave_direction', 'omega', 'freq']
        for icoord in coordinates:
            assert icoord in elevation.coords, f'missing coordinate: {icoord}'

    def test_shape(self, elevation, nfreq, ndir):
        """Test that the elevation dataArray has the correct shape."""
        assert np.squeeze(elevation.values).shape == (nfreq, ndir)

    def test_type(self, elevation):
        """Test that the elevation dataArray has the correct type."""
        assert np.iscomplexobj(elevation)

    def test_empty(self, elevation):
        """Test that the elevation dataArray is initialized with zeros.
        """
        assert np.allclose(np.abs(elevation), 0.0)


class TestRegularWave:
    """Test function :python:`waves.regular_wave`."""

    @pytest.fixture(scope="class")
    def dir(self,):
        """Wave direction [degrees]."""
        return np.random.random()*360

    @pytest.fixture(scope="class")
    def freq(self, f1, nfreq):
        """Wave frequency [Hz]."""
        return f1*np.random.randint(1, nfreq)

    @pytest.fixture(scope="class")
    def amp(self,):
        """Wave amplitude [m]."""
        return 2.5 * np.random.random()

    @pytest.fixture(scope="class")
    def phase(self,):
        """Wave phase [degrees]."""
        return np.random.random() * 360

    @pytest.fixture(scope="class")
    def elevation(self, f1, nfreq, freq, amp, phase, dir):
        """Complex sea state elevation amplitude [m] indexed by
        frequency and direction."""
        return wot.waves.regular_wave(f1, nfreq, freq, amp, phase, dir)

    def test_coordinates(self, elevation):
        """Test that the elevation dataArray has the correct
        coordinates.
        """
        coordinates = ['wave_direction', 'omega', 'freq']
        for icoord in coordinates:
            assert icoord in elevation.coords, f'missing coordinate: {icoord}'

    def test_shape(self, elevation, nfreq, ndir):
        """Test that the elevation dataArray has the correct shape."""
        assert np.squeeze(elevation.values).shape == (nfreq, )

    def test_type(self, elevation):
        """Test that the elevation dataArray has the correct type."""
        assert np.iscomplexobj(elevation)

    def test_direction(self, elevation, dir):
        """Test that the wave direction is correct."""
        dir_out = elevation.wave_direction.values.item()
        assert np.isclose(dir_out, wot.degrees_to_radians(dir))

    def test_value(self, elevation, freq, amp, phase):
        """Test that the complex amplitude [m] has the correct value."""
        amp_out = elevation.sel(omega=freq*2*np.pi).values
        amp_correct = amp*np.exp(1j*wot.degrees_to_radians(phase))
        assert np.isclose(amp_out, amp_correct)

    def test_only_one_value(self, elevation, freq):
        """Test that all other values are zero."""
        # set the single entry to zero
        elev0 = elevation.copy()
        omega = elevation.sel(omega=2*np.pi*freq).omega.values
        idx = np.where(elevation.omega.values == omega)
        elev0.values[idx] = 0
        # check that now the entire array is zero.
        assert np.allclose(elev0.values, 0.0+0.0j)


class TestLongCrestedWave:
    """Test function :python:`waves.long_crested_wave`."""

    @pytest.fixture(scope="class")
    def ndbc_omnidirectional(self,):
        """Omnidirectional spectrum from NDBC data interpolated at
        desired frequencies.
        """
        f1 = 0.02
        time = '2020-01-01T01:40:00.000000000'
        nfreq = 24
        freq = wot.frequency(f1, nfreq, False)
        dir = os.path.join(os.path.dirname(__file__), 'data', 'ndbc')
        spec = ws.read_ndbc_ascii(os.path.join(dir, '41013w2020.txt'))
        return spec.sel(time=time).interp(freq=freq)

    @pytest.fixture(scope="class")
    def nfreq(self, ndbc_omnidirectional):
        """Number of wave frequencies."""
        return len(ndbc_omnidirectional.freq)

    @pytest.fixture(scope="class")
    def ndir(self, ndbc_omnidirectional):
        """Number of wave directions."""
        return len(ndbc_omnidirectional.dir)

    @pytest.fixture(scope="class")
    def direction(self, ndbc_omnidirectional, ndir):
        """Wave direction."""
        return ndbc_omnidirectional.dir.values[np.random.randint(0, ndir)]

    @pytest.fixture(scope="class")
    def elevation(self, ndbc_omnidirectional, direction):
        """Complex sea state elevation amplitude [m] indexed by
        frequency and direction."""
        elev = wot.waves.long_crested_wave(
            ndbc_omnidirectional.efth, direction)
        return elev

    @pytest.fixture(scope="class")
    def pm_f1(self,):
        """Fundamental frequency for the Pierson-Moskowitz spectrum."""
        return 0.05

    @pytest.fixture(scope="class")
    def pm_nfreq(self,):
        """Number of frequencies for the Pierson-Moskowitz spectrum."""
        return 100

    @pytest.fixture(scope="class")
    def pm_hs(self,):
        """Significant wave height for Pierson-Moskowitz spectrum."""
        return 5.0

    @pytest.fixture(scope="class")
    def pm_spectrum(self, pm_f1, pm_nfreq, pm_hs):
        """Pierson-Moskowitz spectrum."""
        Tp = 1.2
        Hs = pm_hs

        def spectrum_func(f):
            return wot.waves.pierson_moskowitz_spectrum(f, fp=1/Tp, hs=Hs)
        spectrum_name = f"Pierson-Moskowitz ({Tp}s, {Hs}m)"

        efth_xr = wot.waves.omnidirectional_spectrum(
            pm_f1, pm_nfreq, spectrum_func, spectrum_name
        )
        return efth_xr

    def test_coordinates(self, elevation):
        """Test that the elevation dataArray has the correct
        coordinates.
        """
        coordinates = ['wave_direction', 'omega', 'freq']
        for icoord in coordinates:
            assert icoord in elevation.coords, f'missing coordinate: {icoord}'

    def test_shape(self, elevation, nfreq, ndir):
        """Test that the elevation dataArray has the correct shape."""
        assert np.squeeze(elevation.values).shape == (nfreq, )

    def test_type(self, elevation):
        """Test that the elevation dataArray has the correct type."""
        assert np.iscomplexobj(elevation)

    def test_direction(self, elevation, direction):
        """Test that the wave direction is correct."""
        dir_out = elevation.wave_direction.values.item()
        assert np.isclose(dir_out, wot.degrees_to_radians(direction))

    def test_spectrum(self, pm_spectrum, pm_hs):
        """Test that the constructed spectrum has the expected Hs."""
        efth = ws.SpecArray(pm_spectrum)
        assert np.isclose(pm_hs, efth.hs().values)

    def test_time_series(self, pm_spectrum, pm_f1, pm_nfreq):
        """Test that the created time series has the desired spectrum."""
        # create time-series
        direction = 0.0
        wave = wot.waves.long_crested_wave(pm_spectrum, direction)
        wave_ts = wot.fd_to_td(wave.values, pm_f1, pm_nfreq, False)
        # calculate the spectrum from the time-series
        t = wot.time(pm_f1, pm_nfreq)
        fs = 1/t[1]
        nnft = len(t)
        [_, S_data] = signal.welch(
            wave_ts.squeeze(), fs=fs, window='boxcar', nperseg=nnft, nfft=nnft,
            noverlap=0
        )
        # check it is equal to the original spectrum
        assert np.allclose(S_data[1:-1], pm_spectrum.values.squeeze()[:-1])


class TestIrregularWave:
    """Test function :python:`waves.irregular_wave`."""

    @pytest.fixture(scope="class")
    def ndbc_spectrum(self,):
        """Spectrum from NDBC data interpolated at desired frequencies.
        """
        f1 = 0.02
        nfreq = 24
        time = '2020-01-01T01:40:00.000000000'
        freq = wot.frequency(f1, nfreq, False)
        markers = ('w', 'd', 'i', 'j', 'k')
        dir = os.path.join(os.path.dirname(__file__), 'data', 'ndbc')
        files = [f'41013{i}2020.txt' for i in markers]
        spec = ws.read_ndbc_ascii([os.path.join(dir, file) for file in files])
        return spec.sel(time=time).interp(freq=freq)

    @pytest.fixture(scope="class")
    def elevation(self, ndbc_spectrum):
        """Complex sea state elevation amplitude [m] indexed by
        frequency and direction."""
        return wot.waves.irregular_wave(ndbc_spectrum.efth)

    def test_coordinates(self, elevation):
        """Test that the elevation dataArray has the correct
        coordinates.
        """
        coordinates = ['wave_direction', 'omega', 'freq']
        for icoord in coordinates:
            assert icoord in elevation.coords, f'missing coordinate: {icoord}'

    def test_shape(self, ndbc_spectrum, elevation):
        """Test that the elevation dataArray has the correct shape."""
        nfreq = len(ndbc_spectrum.freq)
        ndir = len(ndbc_spectrum.dir)
        assert np.squeeze(elevation.values).shape == (nfreq, ndir)

    def test_type(self, elevation):
        """Test that the elevation dataArray has the correct type."""
        assert np.iscomplexobj(elevation)


class TestRandomPhase:
    """Test function :python:`waves.random_phase`."""

    @pytest.fixture(scope="class")
    def shape(self,):
        """Shape of the phase matrix, randomized each time the test is
        run.
        """
        return (np.random.randint(10, 100), np.random.randint(10, 100))

    @pytest.fixture(scope="class")
    def phase_mat(self, shape):
        """Random phase matrix."""
        return wot.waves.random_phase(shape)

    def test_mat_shape(self, phase_mat, shape):
        """Test correct shape of random phase matrix."""
        assert phase_mat.shape == shape

    def test_mat_values(self, phase_mat):
        """Test all phases are between [-pi, pi])."""
        assert (np.max(phase_mat) < np.pi) and (np.min(phase_mat) >= -np.pi)

    def test_float(self,):
        """Test that the function works for a single float instead of a
        matrix.
        """
        phase = wot.waves.random_phase()
        assert (phase < np.pi) and (phase >= -np.pi)


# TODO: Move everything below to wavespectra.construct
class TestWaveSpectra:
    def test_omnidirectional_spectrum(self, f1, nfreq, fp, hs):
        def spectrum_func(
            f): return wot.waves.pierson_moskowitz_spectrum(f, fp, hs)
        wave_spec = wot.waves.omnidirectional_spectrum(
            f1, nfreq, spectrum_func, "Pierson-Moskowitz")

        # the values should be the same as calling the spectrum function
        freq = wot.frequency(f1, nfreq, False)
        spec_test = spectrum_func(freq)

        assert np.allclose(spec_test, wave_spec.values.flatten())

    def test_spectrum(self, f1, nfreq, fp, hs, ndir):
        s_max = 10
        directions = np.linspace(0, 360, ndir, endpoint=False)
        dm = directions[np.random.randint(0, ndir)]

        def spectrum_func(f): 
            return wot.waves.pierson_moskowitz_spectrum(f, fp, hs)

        def spread_func(f, d): 
            return wot.waves.spread_cos2s(f, d, dm, fp, s_max)
        spectrum_name, spread_name = "Pierson-Moskowitz", "Cos2s"
        wave_spec = wot.waves.spectrum(
            f1, nfreq, directions, spectrum_func, spread_func,
            spectrum_name, spread_name)

        # integral over all angles should be equal to omnidirectional
        spec_omni = wot.waves.omnidirectional_spectrum(
            f1, nfreq, spectrum_func, spectrum_name)
        spec_omni = spec_omni.values.flatten()
        ddir = (wave_spec.dir[1] - wave_spec.dir[0]).values
        integral_d = wave_spec.sum(dim='dir').values * ddir

        # mean direction
        dfreq = (wave_spec.freq[1] - wave_spec.freq[0]).values
        integral_f = wave_spec.sum(dim='freq').values * dfreq

        assert wave_spec.shape == (nfreq, ndir)  # shape
        assert np.allclose(integral_d, spec_omni, rtol=0.01)
        assert directions[np.argmax(integral_f)] == dm

    def test_pierson_moskowitz_spectrum(self, f1, nfreq, fp, hs):
        spectrum = wot.waves.pierson_moskowitz_spectrum

        # scalar
        freq_1 = 0.4
        spec1 = spectrum(freq_1, fp, hs)

        # vector
        freqs = wot.frequency(f1, nfreq, False)
        spec = spectrum(freqs, fp, hs)

        # total elevation variance
        freqs_int = np.linspace(0, 10, 1000)[1:]
        total_variance_calc = np.trapz(spectrum(freqs_int, fp, hs), freqs_int)
        a_param, b_param = wot.waves.pierson_moskowitz_params(fp, hs)
        total_variance_theory = a_param/(4*b_param)

        assert isinstance(spec1, float)  # scalar
        assert spec.shape == freqs.shape  # vector shape
        assert np.isclose(total_variance_calc,
                          total_variance_theory)  # integral

    def test_jonswap_spectrum(self, f1, nfreq, fp, hs):
        spectrum = wot.waves.jonswap_spectrum

        # scalar
        freq_1 = 0.4
        spec1 = spectrum(freq_1, fp, hs)

        # vector
        freqs = wot.frequency(f1, nfreq, False)
        spec = spectrum(freqs, fp, hs)

        # reduces to PM
        spec_gamma1 = spectrum(freqs, fp, hs, gamma=1.0)
        spec_pm = wot.waves.pierson_moskowitz_spectrum(freqs, fp, hs)

        assert isinstance(spec1, float)  # scalar
        assert spec.shape == freqs.shape  # vector shape
        assert np.allclose(spec_gamma1, spec_pm)  # reduces to PM

    def test_spread_cos2s(self, f1, nfreq, fp, ndir):
        """Confirm that energy is spread correctly accross wave directions.
        Integral over all directions of the spread function gives (vector)
        1.
        """
        directions = np.linspace(0, 360, ndir, endpoint=False)
        wdir_mean = directions[np.random.randint(0, ndir)]
        freqs = wot.frequency(f1, nfreq, False)
        s_max = 10
        spread = wot.waves.spread_cos2s(freq=freqs,
                                        directions=directions,
                                        dm=wdir_mean,
                                        fp=fp,
                                        s_max=s_max)
        ddir = directions[1]-directions[0]
        dfreq = freqs[1] - freqs[0]
        integral_d = np.sum(spread, axis=1)*ddir
        integral_f = np.sum(spread, axis=0)*dfreq

        assert directions[np.argmax(integral_f)] == wdir_mean  # mean dir
        assert np.allclose(integral_d, np.ones(
            (1, nfreq)), rtol=0.01)  # omnidir

    def test_general_spectrum(self, f1, nfreq):
        freq = wot.frequency(f1, nfreq, False)
        a_param = np.random.random()*10
        b_param = np.random.random()*10
        spec_f1 = wot.waves.general_spectrum(a_param, b_param, 1.0)
        spec_a0 = wot.waves.general_spectrum(0, b_param, freq)
        spec_b0 = wot.waves.general_spectrum(a_param, 0, freq)

        a_vec = np.random.random(freq.shape)*10
        b_vec = np.random.random(freq.shape)*10
        spec_vec = wot.waves.general_spectrum(a_vec, b_vec, freq)

        # types and shapes
        assert isinstance(spec_f1, float)
        assert spec_a0.shape == spec_b0.shape == freq.shape
        assert spec_vec.shape == freq.shape
        # values
        assert np.isclose(spec_f1, a_param * np.exp(-b_param))
        assert np.allclose(spec_a0, 0.0)
        assert np.allclose(spec_b0, a_param * freq**(-5))

    def test_pierson_moskowitz_params(self, fp, hs):
        params = wot.waves.pierson_moskowitz_params(fp, hs)

        assert len(params) == 2  # returns two floats
        for iparam in params:
            assert isinstance(iparam, float)



"""Provide the wave definition structure and provide
functions for creating different common types of waves.

This module provides the (empty) data structure for waves in
:python:`wecopttool`.
It also provides functions for creating common types of waves such as
regular waves and irregular waves.
The data structure is a 2D complex :py:module:xarray.DataArray
containing the complex amplitude.
The 2D coordinates are: wave angular frequency :python:`omega` (rad/s)
and direction :python:`wave_direction` (rad).

This module uses wave spectrum data in the
:py:class:`wavespectra.SpecArray` format.
"""


from __future__ import annotations


__all__ = [
    "elevation_fd",
    "regular_wave",
    "long_crested_wave",
    "irregular_wave",
    "random_phase",
    "omnidirectional_spectrum",
    "spectrum",
    "pierson_moskowitz_spectrum",  # TODO: Remove all below after wavespectra
    "jonswap_spectrum",
    "spread_cos2s",
    "general_spectrum",
    "pierson_moskowitz_params",
]


import logging
from typing import Callable, Mapping, Union, Optional, Iterable

import numpy as np
from numpy.typing import ArrayLike
from numpy import ndarray
from xarray import DataArray
from scipy.special import gamma

from wecopttool.core import frequency, degrees_to_radians, frequency_parameters


# logger
_log = logging.getLogger(__name__)


def elevation_fd(
    f1: float,
    nfreq: int,
    directions: Union[float, ArrayLike],
    amplitudes: Optional[ArrayLike] = None,
    phases: Optional[ArrayLike] = None,
    attr: Optional[Mapping] = None,
    seed: Optional[float] = None,
) -> DataArray:
    """Construct the complex wave elevation
    :py:class:`xarray.DataArray`.

    This is the complex wave elevation (m) indexed by radial frequency
    (rad/s) and wave direction (rad).
    The coordinate units and names match those from *Capytaine*.

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [Hz].
    nfreq
        Number of frequencies (not including zero frequency),
        i.e., :python:`freq = [0, f1, 2*f1, ..., nfreq*f1]`.
    directions
        Wave directions in degrees. 1D array.
    amplitudes:
        Wave elevation amplitude in meters.
    phases:
        Wave phases in degrees.
    attr:
        Additional attributes (metadata) to include in the
        :py:class:`xarray.DataArray`.
    seed
        Seed for random number generator. Used for reproducibility.
        Generally should not be used except for testing.
    """
    directions = np.atleast_1d(degrees_to_radians(directions, sort=False))
    ndirections = len(directions)
    freq = frequency(f1, nfreq, False)
    omega = freq*2*np.pi

    dims = ('omega', 'wave_direction')
    omega_attr = {'long_name': 'Radial frequency', 'units': 'rad/s'}
    freq_attr = {'long_name': 'Frequency', 'units': 'Hz'}
    dir_attr = {'long_name': 'Wave direction', 'units': 'rad'}
    coords = {'omega': (dims[0], omega, omega_attr),
              'freq': (dims[0], freq, freq_attr),
              'wave_direction': (dims[1], directions, dir_attr)}

    if amplitudes is None:
        amplitudes = np.zeros([nfreq, ndirections])

    if phases is None:
        phases = random_phase([nfreq, ndirections],seed)
    else:
        phases = degrees_to_radians(phases, False)

    camplitude = amplitudes * np.exp(1j*phases)

    attr = {} if attr is None else attr
    attrs = {'units': 'm', 'long_name': 'Wave elevation'} | attr

    waves = DataArray(camplitude, dims=dims, coords=coords,
                         attrs=attrs, name='wave_elev')

    return waves.sortby(waves.wave_direction)


def regular_wave(
    f1: float,
    nfreq: int,
    freq: float,
    amplitude: float,
    phase: Optional[float] = None,
    direction: float = 0.0,
) -> DataArray:
    """Create the dataset for a regular wave.

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [Hz].
    nfreq
        Number of frequencies (not including zero frequency),
        i.e., :python:`freq = [0, f1, 2*f1, ..., nfreq*f1]`.
    freq
        Frequency (in Hz) of the regular wave. If :python:`freq` is not
        in the frequency array, the closest value is used and a warning
        is displayed.
    amplitude
        Amplitude (in meters) of the regular wave.
    phase
        Phase (in degrees) of the regular wave.
    direction
        Direction (in degrees) of the regular wave.
    """

    # attributes & index
    omega = freq*2*np.pi
    tmp_waves = elevation_fd(f1, nfreq, direction)
    iomega = tmp_waves.sel(omega=omega, method='nearest').omega.values
    ifreq = iomega/(2*np.pi)

    if not np.isclose(iomega, omega):
        _log.warning(
            f"Requested frequency {freq} Hz is not in array. " +
            f"Using nearest value of {ifreq} Hz."
        )

    attrs = {
        'Wave type': 'Regular',
        'Frequency (Hz)': ifreq,
        'Amplitude (m)': amplitude,
        'Phase (degrees)': phase,
        'Direction (degrees)': direction,
    }

    # phase
    if phase is None:
        rphase = random_phase()
        phase = np.degrees(rphase)
    else:
        rphase = degrees_to_radians(phase)

    # wave elevation
    tmp = np.zeros([nfreq, 1])
    waves = elevation_fd(f1, nfreq, direction, tmp, tmp, attrs)
    waves.loc[{'omega': iomega}] = amplitude  * np.exp(1j*rphase)

    return waves


def long_crested_wave(
    efth: DataArray,
    direction: Optional[float] = 0.0,
    seed: Optional[float] = None,
) -> DataArray:
    """Create a complex frequency-domain wave elevation from an
    omnidirectional spectrum.

    The omnidirectional spectrum is in the
    :py:class:`wavespectra.SpecArray` format.

    .. note:: The frequencies must be evenly-spaced with spacing equal
              to the first frequency. This is not always the case when
              e.g. reading from buoy data. Use interpolation as
              :python:`da.interp(freq=[...])`.

    Parameters
    ----------
    efth
        Omnidirection wave spectrum in units of m^2/Hz, in the format
        used by :py:class:`wavespectra.SpecArray`.
    direction
        Direction (in degrees) of the long-crested wave.
    seed
        Seed for random number generator. Used for reproducibility.
        Generally should not be used except for testing.
    """
    f1, nfreq = frequency_parameters(efth.freq.values, False)
    df = f1

    values = efth.values
    values[values<0] = np.nan
    amplitudes = np.sqrt(2 * values * df)

    attr = {
        'Wave type': 'Long-crested irregular',
        'Direction (degrees)': direction,
    }

    return elevation_fd(f1, nfreq, direction, amplitudes, None, attr, seed)


def irregular_wave(efth: DataArray,
                   seed: Optional[float] = None,) -> DataArray:
    """Create a complex frequency-domain wave elevation from a spectrum.

    The omnidirectional spectrum is in the
    :py:class:`wavespectra.SpecArray` format.

    .. note:: The frequencies must be evenly-spaced with spacing equal
              to the first frequency. This is not always the case when
              e.g. reading from buoy data. Use interpolation as
              :python:`da.interp(freq=[...])`.

    .. note:: The wave directions must also be evenly spaced.

    Parameters
    ----------
    efth
        Wave spectrum in units of m^2/Hz/deg, in the format used by
        :py:class:`wavespectra.SpecArray`.
    seed
        Seed for random number generator. Used for reproducibility.
        Generally should not be used except for testing.
    """
    f1, nfreq = frequency_parameters(efth.freq.values, False)
    directions = efth.dir.values
    df = f1
    dd = np.sort(directions)[1]-np.sort(directions)[0]

    values = efth.values
    values[values<0] = np.nan
    amplitudes = np.sqrt(2 * values * df * dd)

    attr = {'Wave type': 'Irregular'}

    return elevation_fd(f1, nfreq, directions, amplitudes, None, attr, seed)


def random_phase(
    shape: Optional[Union[Iterable[int], int]] = None,
    seed: Optional[float] = None,
) -> Union[float , ndarray]:
    """Generate random phases in range [-π, π) radians.

    Parameters
    ----------
    shape
        Shape of the output array of random phases.
    seed
        Seed for random number generator. Used for reproducibility.
        Generally should not be used except for testing.
    """
    rng = np.random.default_rng(seed)
    return rng.random(shape)*2*np.pi - np.pi


# TODO: Move everything below to wavespectra.construct
#       wavespectra is good at reading from simulations or measurements
#       but not good yet at constructing from parametric models.
def omnidirectional_spectrum(
    f1: float,
    nfreq: int,
    spectrum_func: Callable,
    spectrum_name: str = '',
) -> DataArray:
    """Create the dataset for an omnidirectional wave spectrum in the
    :py:class:`wavespectra.SpecArray` format.

    Examples
    --------
    Define wave parameters.

    >>> from wecopttool.waves import omnidirectional_spectrum as omnispec
    >>> from wecopttool.waves import pierson_moskowitz_spectrum as pm
    >>> Hs = 5
    >>> Tp = 6
    >>> fp = 1/Tp

    Generate the wave using a Pierson-Moskowitz idealized spectrum.

    >>> omnidirectional_spectrum = omnispec(
    ...     f1=fp/10,
    ...     nfreq=30,
    ...     spectrum_func=lambda f: pm(freq=f, fp=fp, hs=Hs),
    ...     spectrum_name="Pierson-Moskowitz",
    ...     )

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [Hz].
    nfreq
        Number of frequencies (not including zero frequency),
        i.e., :python:`freq = [0, f1, 2*f1, ..., nfreq*f1]`.
    spectrum_func
        Wave spectrum function. Maps frequencies to amplitude spectrum.
        :python:`Union[float, ArrayLike] -> Union[float, ndarray]`.
        Units: :math:`m^2/Hz`.
    spectrum_name
        Name of the spectrum function.
    """
    # dimensions & coordinates
    freq = frequency(f1, nfreq, False)
    dims = ('freq', 'dir')
    freq_attr = {
        'long_name': 'wave frequency',
        'units': 'Hz'
    }
    dir_attr = {
        'long_name': 'wave direction',
        'units': 'deg'
    }
    dir = [0.0,]
    coords = [(dims[0], freq, freq_attr), (dims[1], dir, dir_attr)]

    # spectrum
    efth = spectrum_func(freq).reshape(nfreq, 1)
    efth_attr = {
        'long_name': 'omnidirectional spectrum',
        'units': 'm^2/Hz',
        'Omnidirectional Spectrum': spectrum_name,
    }

    return DataArray(efth, dims=dims, coords=coords, attrs=efth_attr)


def spectrum(
    f1: float,
    nfreq: int,
    directions: Union[float, ArrayLike],
    spectrum_func: Callable,
    spread_func: Callable,
    spectrum_name: str = '',
    spread_name: str = '',
) -> DataArray:
    """Create the dataset for an irregular wave in the
    :py:class:`wavespectra.SpecArray` format.

    Examples
    --------
    Define the desired spectrum parameters.

    >>> import wecopttool as wot
    >>> import numpy as np
    >>> Hs = 5
    >>> Tp = 6
    >>> fp = 1/Tp
    >>> directions = np.linspace(0, 360, 36, endpoint=False)

    Create a function handle to define the spectral density,

    >>> def spectrum_func(f):
    ...    return wot.waves.pierson_moskowitz_spectrum(freq=f,
    ...                                                fp=fp,
    ...                                                hs=Hs)

    and a spreading function handle for spreading.

    >>> def spread_func(f, d):
    ...     return wot.waves.spread_cos2s(freq=f,
    ...                                   directions=d,
    ...                                   dm=10,
    ...                                   fp=fp,
    ...                                   s_max=10)

    Generate the spectrum.

    >>> wave = wot.waves.spectrum(f1=fp/10,
    ...                           nfreq=20,
    ...                           directions=directions,
    ...                           spectrum_func=spectrum_func,
    ...                           spread_func=spread_func,
    ...                           spectrum_name="Pierson-Moskowitz",
    ...                           spread_name="cosine-2s",)

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [Hz].
    nfreq
        Number of frequencies (not including zero frequency),
        i.e., :python:`freq = [0, f1, 2*f1, ..., nfreq*f1]`.
    directions
        Wave directions in degrees. 1D array, evenly spaced.
    spectrum_func
        Wave spectrum function. Maps frequencies to amplitude spectrum.
        :python:`Union[float, ArrayLike] -> Union[float, np.ndarray]`.
        Units: :math:`m^2/Hz`.
    spread_func
        Wave spreading function.
        Maps wave frequencies and directions to spread value.
        :python:`tuple[ Union[float, ArrayLike], Union[float, ArrayLike]] -> ndarray`.
        Units: :math:`1/degree`.
    spectrum_name
        Name of the spectrum function.
    spread_name
        Name of the spread function.
    """
    # dimensions & coordinates
    freq = frequency(f1, nfreq, False)
    dims = ('freq', 'dir')
    freq_attr = {
        'long_name': 'wave frequency',
        'units': 'Hz'
    }
    dir_attr = {
        'long_name': 'wave direction',
        'units': 'deg'
    }
    dir = directions
    coords = [(dims[0], freq, freq_attr), (dims[1], dir, dir_attr)]

    # spectrum
    efth = spectrum_func(freq).reshape(nfreq, 1) * spread_func(freq, dir)
    efth_attr = {
        'long_name': 'spectrum',
        'units': 'm^2/Hz/deg',
        'Omnidirectional Spectrum': spectrum_name,
        'Spreading function': spread_name,
    }

    return DataArray(efth, dims=dims, coords=coords, attrs=efth_attr)


def pierson_moskowitz_spectrum(
    freq: Union[float, ArrayLike],
    fp: float,
    hs: float,
) -> Union[float, ndarray]:
    """Calculate the Pierson-Moskowitz omni-directional wave spectrum
    for the specified frequencies  and parameters.

    This is included as one example of a spectrum function.

    Return is in units of :math:`m^2/Hz`.

    Parameters
    ----------
    freq
        Wave frequencies.
    fp
        Peak frequency of the sea-state in :math:`Hz`.
    hs
        Significant wave height of the sea-state in :math:`m`.
    """
    a_param, b_param = pierson_moskowitz_params(fp, hs)
    return general_spectrum(a_param, b_param, freq)


def jonswap_spectrum(
    freq: Union[float, ArrayLike],
    fp: float,
    hs: float,
    gamma: float = 3.3,
) -> Union[float, ndarray]:
    """Calculate the Joint North Sea Wave Project (JONSWAP)
    omni-directional wave spectrum for the specified frequencies and
    parameters.

    See, e.g., :title:`DNV-RP-C205`

    For :python:`gamma = 1`, the JONSWAP spectrum reduces to a
    Pierson-Moskowitz spectrum.

    Return is in units of :math:`m^2/Hz`.

    Parameters
    ----------
    freq
        Wave frequencies in :math:`Hz`.
    fp
        Peak frequency in :math:`Hz`.
    hs
        Significant wave height in :math:`m`.
    gamma
        Peakedness factor.
    """
    # Pierson-Moskowitz parameters
    a_param_pm, b_param = pierson_moskowitz_params(fp, hs)

    # JONSWAP parameters
    sigma_a = 0.07
    sigma_b = 0.09
    sigma = np.piecewise(freq,
                         condlist=[freq <= fp, freq > fp],
                         funclist=[sigma_a, sigma_b])
    alpha = np.exp(-1*((freq/fp - 1)/(np.sqrt(2)*sigma))**2)
    c_param = 1-0.287*np.log(gamma)
    a_param = a_param_pm * c_param * gamma**alpha

    return general_spectrum(a_param, b_param, freq)


def spread_cos2s(
    freq: Union[float, ArrayLike],
    directions: Union[float, ArrayLike],
    dm: float,
    fp: float,
    s_max: float,
) -> Union[float, np.ndarray]:
    """Calculate the Cosine-2s spreading function for the specified
    frequencies and wave directions.

    This is included as one example of a spreading function.

    Return is in units of :math:`1/degrees`.

    Parameters
    ----------
    freq
        Wave frequencies in Hz.
    directions
        Wave directions relative to mean/wind direction in degrees.
    dm: float
        Mean wave direction in degrees.
    fp
        Peak frequency of the sea-state in :math:`Hz`.
    s_max
        The spreading parameter. Larger values corresponds to less
        spread. For fully developed seas a value of 10 is a good choice.
    """
    freq = np.atleast_1d(freq)
    rdir = degrees_to_radians(directions-dm, False)
    pow = np.ones(len(freq)) * 5.0
    pow[freq > fp] = -2.5
    s_param = s_max * (freq/fp)**pow
    cs = 2**(2*s_param-1)/np.pi * (gamma(s_param+1))**2/gamma(2*s_param+1)
    return np.pi/180 * (cs * np.power.outer(np.cos(rdir/2), 2*s_param)).T


def general_spectrum(
    a_param: Union[float, ArrayLike],
    b_param: Union[float, ArrayLike],
    freq: Union[float, ArrayLike],
) -> Union[float, ArrayLike]:
    """Create a spectrum function.

    The general omni-directional spectrum formulation is
    :math:`S(f) = A f^(-5) e^(-B f^(-4))`.

    Return is in units of :math:`m^2/Hz`.

    Parameters
    ----------
    a_param
        Parameter :math:`A` in the general spectrum equation.
    b_param
        Parameter :math:`B` in the general spectrum equation.
    freq
        Wave frequencies in Hz.
    """
    spectrum = a_param * freq**(-5) * np.exp(-b_param * freq**(-4))
    # if scalar, return scalar
    spectrum = spectrum.item() if (spectrum.size == 1) else spectrum
    return spectrum


def pierson_moskowitz_params(fp: float, hs: float) -> Union[float, ndarray]:
    """Return the two PM parameters for the general spectrum formulation.

    Parameters
    ----------
    fp
        Peak frequency in :math:`Hz`.
    hs
        Significant wave height in :math:`m`.

    Returns
    -------
    a_param
        Parameter :math:`A` in the general spectrum equation.
    b_param
        Parameter :math:`B` in the general spectrum equation.

    See Also
    --------
    general_spectrum,
    """
    b_param = (1.057*fp)**4
    a_param = hs**2 / 4 * b_param
    return a_param, b_param



""" Unit tests for functions in the :python:`hydrostatics.py` module.
"""

import pytest
import capytaine as cpy
import numpy as np

import wecopttool as wot
from wecopttool.core import _default_parameters


# setup simple constant density rectangular barge
@pytest.fixture()
def rho():
    """Water density [kg/m^3]."""
    return _default_parameters['rho']


@pytest.fixture()
def g():
    """Gravitational acceleration [m/s^2]."""
    return _default_parameters['g']


@pytest.fixture()
def lx():
    """Barge length [m]."""
    return 4.0


@pytest.fixture()
def ly():
    """Barge width [m]."""
    return 5.0


@pytest.fixture()
def lz():
    """Barge height [m]."""
    return 2.0


@pytest.fixture()
def cog():
    """Center of gravity of the barge."""
    return (0.0, 0.0, 0.0)


@pytest.fixture()
def mass(lx, ly, lz, rho):
    """Mass of the barge."""
    return lx*ly*lz/2 * rho


@pytest.fixture()
def fb(lx, ly, lz):
    """Simple constant density rectangular barge."""
    rect = cpy.RectangularParallelepiped(
        (lx, ly, lz), resolution=(100, 100, 10), center=(0.0, 0.0, 0.0,))
    rect.add_all_rigid_body_dofs()
    rect.center_of_mass = None
    fb.mass = None
    return rect


@pytest.fixture()
def fb_mass(fb, cog, mass):
    """Simple constant density rectangular barge with the mass
    properties specified.
    """
    fbc = fb.copy()
    fbc.center_of_mass = cog
    fbc.mass = mass
    return fbc


@pytest.fixture()
def stiffness(lx, ly, lz, rho, g):
    """True (theoretical/calculated) value of the stiffness matrix for
    the barge.
    """
    stiffness = np.zeros([6, 6])
    stiffness[2, 2] = rho*g * lx*ly
    stiffness[2, 3] = 0.0
    stiffness[2, 4] = 0.0
    stiffness[3, 3] = (rho*g * lx*ly**3 /12) + rho*g*(lx*ly*lz/2)*(-lz/4)
    stiffness[3, 4] = 0.0
    stiffness[3, 5] = 0.0
    stiffness[4, 4] = (rho*g * ly*lx**3 /12) + rho*g*(lx*ly*lz/2)*(-lz/4)
    stiffness[4, 5] = 0.0
    stiffness[3, 2] = stiffness[2, 3]
    stiffness[4, 2] = stiffness[2, 4]
    stiffness[4, 3] = stiffness[3, 4]
    return stiffness


@pytest.fixture()
def inertia(lx, ly, lz, mass):
    """True (theoretical/calculated) value of the inertia matrix for
    the barge.
    """
    inertia = np.zeros([6, 6])
    inertia[0, 0] = mass
    inertia[1, 1] = mass
    inertia[2, 2] = mass
    inertia[3, 3] = mass/12 * (ly**2 + lz**2)
    inertia[3, 4] = 0.0
    inertia[3, 5] = 0.0
    inertia[4, 3] = 0.0
    inertia[4, 4] = mass/12 * (lx**2 + lz**2)
    inertia[4, 5] = 0.0
    inertia[5, 3] = 0.0
    inertia[5, 4] = 0.0
    inertia[5, 5] = mass/12 * (lx**2 + ly**2)
    return inertia


class TestStiffnessMatrix:
    """Test function :python:`hydrostatics.stiffness_matrix`."""

    def test_inferred_cog(self, fb, rho, g, stiffness):
        """Test the function with the center of gravity not provided,
        but inferred.
        """
        stiffness_calc = wot.hydrostatics.stiffness_matrix(fb, rho, g)
        assert np.allclose(stiffness, stiffness_calc, rtol=0.01)


    def test_given_cog(self, fb, rho, g, cog, stiffness):
        """Test the function with the center of gravity provided as a
        function input and not in the floating body.
        """
        stiffness_calc = wot.hydrostatics.stiffness_matrix(fb, rho, g, cog)
        assert np.allclose(stiffness, stiffness_calc, rtol=0.01)


    def test_given_cog_fb(self, fb_mass, rho, g, stiffness):
        """Test the function with the center of gravity provided in the
        floating body and not as a function input.
        """
        stiffness_calc = wot.hydrostatics.stiffness_matrix(fb_mass, rho, g)
        assert np.allclose(stiffness, stiffness_calc, rtol=0.01)


    def test_given_cog_redundant(self, fb_mass, rho, g, cog, stiffness):
        """Test the function with the center of gravity provided in both
        the floating body and the function inputs.
        """
        stiffness_calc = wot.hydrostatics.stiffness_matrix(fb_mass, rho, g, cog)
        assert np.allclose(stiffness, stiffness_calc, rtol=0.01)


    def test_cog_mismatch(self, fb_mass, rho, g):
        """Test that the function fails if the center of gravity is
        provided in both the floating body and the function inputs but
        have different values.
        """
        cog_wrong = (0, 0, -0.1)
        with pytest.raises(ValueError):
            wot.hydrostatics.stiffness_matrix(fb_mass, rho, g, cog_wrong)


class TestInertiaMatrix:
    """Test function :python:`hydrostatics.test_inertia_matrix`."""

    def test_inferred_mass(self, fb, rho, inertia):
        """Test the function with the mass not provided, but inferred.
        """
        inertia_calc = wot.hydrostatics.inertia_matrix(fb, rho)
        assert np.allclose(inertia, inertia_calc, rtol=0.01)


    def test_given_mass(self, fb, rho, cog, mass, inertia):
        """Test the function with the mass provided as a function input
        and not in the floating body.
        """
        inertia_calc = wot.hydrostatics.inertia_matrix(fb, rho, cog, mass)
        assert np.allclose(inertia, inertia_calc, rtol=0.01)


    def test_given_mass_fb(self, fb_mass, rho, cog, mass, inertia):
        """Test the function with the mass provided in the floating body
        and not as a function input.
        """
        inertia_calc = wot.hydrostatics.inertia_matrix(fb_mass, rho)
        assert np.allclose(inertia, inertia_calc, rtol=0.01)


    def test_given_mass_redundant(self, fb_mass, rho, cog, mass, inertia):
        """Test the function with the mass provided in both the floating
        body and the function inputs.
        """
        inertia_calc = wot.hydrostatics.inertia_matrix(fb_mass, rho, cog, mass)
        assert np.allclose(inertia, inertia_calc, rtol=0.01)


    def test_mass_mismatch(self, fb_mass, rho, cog, mass):
        """Test that the function fails if the mass is provided in both
        the floating body and the function inputs but have different
        values.
        """
        mass_wrong = 0.9*mass
        with pytest.raises(ValueError):
            wot.hydrostatics.inertia_matrix(fb_mass, rho, cog, mass_wrong)
